# pylint: disable=too-many-lines
from __future__ import annotations

import uuid
import re

from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Callable, Generator, Type, Iterable, Literal, TYPE_CHECKING

from slate import Event, Span, Terminal, Key
from zenith.markup import zml_pre_process, zml_get_spans, FULL_RESET

from ..enums import Alignment, Overflow, MouseAction, Positioning
from ..frames import Frame, get_frame
from ..style_map import StyleMap
from ..state_machine import StateMachine

if TYPE_CHECKING:
    from ..application import Application, Page

__all__ = ["Widget", "widget_types", "handle_mouse_on_children"]

RE_FULL_UNSETTER = re.compile(r"\/(?=\]| )")


@dataclass
class BoundStyle:
    """A callable object containing both style and fill markup."""

    style: str
    fill: str

    def __call__(self, item: str) -> str:
        return f"[{self.fill}{self.style}]{item}"


def _compute(spec: int | float | None, hint: int) -> int:
    if isinstance(spec, float):
        return int(spec * hint)

    # Auto (fill available)
    if spec is None:
        return hint

    # Static (int)
    return spec


def _build_scrollbar(
    rail: str, thumb: str, size: int, current: float, ratio: float
) -> Generator[str, None, None]:
    """Builds the characters for a scrollbar.

    Args:
        rail: The character used for the 'rail' of the scrollbar.
        thumb: The character used for the 'thumb' of the scrollbar.
        size: The total available size for the scrollbar to follow.
        current: The current value/position of the scrollbar, 0-1.
        ratio: The visible ('viewport') size divided by the total ('virtual') size.

    Yields:
        `rail` or `thumb`, depending on which character comes next, a total of `size`
        times.
    """

    thumb_length = int(size * ratio)
    move_area = size - thumb_length

    thumb_center = int(current * move_area)
    half_thumb, offset = divmod(thumb_length, 2)
    thumb_center += half_thumb

    thumb_start = thumb_center - half_thumb
    thumb_end = thumb_center + half_thumb + offset

    if thumb_end >= size - 1:
        thumb_end = size
        thumb_start = thumb_end - thumb_length - offset

    for i in range(size):
        if thumb_start <= i <= thumb_end:
            yield thumb
            continue

        yield rail


def _overflows(real: int, virt: int) -> bool:
    """Determines whether the given real and virtual dimensions overflow."""

    return real / max(virt, 1) <= 1.0


class Widget:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """This is a docstring."""

    width: int | float | None
    """The hint used by the widget to calculate it's width. See dimension hints."""

    height: int | float | None
    """The hint used by the widget to calculate it's height. See dimension hints."""

    computed_width: int
    """The current width of the widget, calculated from `width`'s hint."""

    computed_height: int
    """The current height of the widget, calculated from `height`'s hint."""

    scroll: tuple[int, int]
    """The widget's (horizontal, vertical) scrolling offset."""

    position: tuple[int, int]
    """The widget's (horizontal, vertical) position."""

    app: "Application"
    """The most recently run Application instance."""

    scroll_step: int = 1

    state_machine = StateMachine(
        states=("idle", "hover", "selected", "active"),
        transitions={
            "idle": {
                "HOVERED": "hover",
                "SELECTED": "selected",
                "CLICKED": "active",
            },
            "hover": {
                "RELEASED": "idle",
                "SELECTED": "selected",
                "CLICKED": "active",
            },
            "selected": {
                "UNSELECTED": "idle",
                "CLICKED": "active",
            },
            "active": {
                "RELEASED": "hover",
                "RELEASED_KEYBOARD": "selected",
            },
            "/": {
                "SUBSTATE_ENTER_SCROLLING_X": "/scrolling_x",
                "SUBSTATE_ENTER_SCROLLING_Y": "/scrolling_y",
            },
            "/scrolling_x": {
                "SUBSTATE_EXIT_SCROLLING_X": "/",
            },
            "/scrolling_y": {
                "SUBSTATE_EXIT_SCROLLING_Y": "/",
            },
        },
    )
    """The state machine keeps track of the current state of the widget.

    It has a tuple of possible states, and a mapping of:

        {current_state: {action: resulting_state, ...}

    ...that are used for state-transitions.
    """

    style_map = StyleMap(
        {
            "idle": {
                "fill": "@ui.panel1-3",
                "frame": "ui.panel1+1",
                "content": "ui.text-1",
                "scrollbar_x": "ui.panel1+3*0.2",
                "scrollbar_y": "ui.panel1+3*0.2",
            },
            "hover": {
                "fill": "@ui.panel1-3",
                "frame": "ui.panel1+1",
                "content": "ui.text-1",
                "scrollbar_x": "ui.panel1+3*0.2",
                "scrollbar_y": "ui.panel1+3*0.2",
            },
            "selected": {
                "fill": "@ui.panel1-3",
                "frame": "ui.panel1+1",
                "content": "ui.text+1",
                "scrollbar_x": "ui.panel1+3*0.2",
                "scrollbar_y": "ui.panel1+3*0.2",
            },
            "active": {
                "fill": "@ui.panel1-3",
                "frame": "ui.panel1+1",
                "content": "ui.text-1",
                "scrollbar_x": "ui.panel1+3*0.2",
                "scrollbar_y": "ui.panel1+3*0.2",
            },
            "/scrolling_x": {
                "scrollbar_x": "ui.primary",
            },
            "/scrolling_y": {
                "scrollbar_y": "ui.primary",
            },
        }
    )
    """The style map is the lookup table for the widget's styles at certain states."""

    rules = ""
    """YAML rules for this widget, applied only when added as part of an Application."""

    _frame: Frame

    def __init__(
        self,
        eid: str | None = None,
        group: str | None = None,
        groups: tuple[str, ...] = tuple(),
    ) -> None:
        """Initializes a Widget.

        Args:
            eid: The id for this widget. Defaults to a UUID.
            group: If set, `groups` is overwritten as `(group,)`.
            groups: The initial groups this widget will belong to.
        """

        self.eid = eid or str(uuid.uuid4())
        if group is not None:
            groups = (group,)
        self.groups = groups
        self.scroll = (0, 0)
        self.position = (0, 0)
        self.state_machine = deepcopy(self.state_machine)
        self.parent: Widget | "Page" | None = None

        self.width: int | float | None = None
        self.height: int | float | None = None
        # These conversions are handled in their properties
        self.frame = get_frame(None)  # type: ignore
        self.alignment = ("start", "start")  # type: ignore
        self.overflow = ("hide", "hide")  # type: ignore
        self.positioning = Positioning.DYNAMIC

        self._virtual_width = 0
        self._virtual_height = 0
        self._last_query: str | None = None
        self._selected_index: int | None = None

        self._clip_start: int | None = None
        self._clip_end: int | None = None
        self._bindings: dict[str, Event] = {}

        self.computed_width = 1
        self.computed_height = 1

        self.setup()

        widget_types[type(self).__name__] = type(self)

    @property
    def terminal(self) -> Terminal | None:
        """Returns the app's terminal."""

        if self.app is None:
            return None

        return self.app.terminal

    @property
    def state(self) -> str:
        """Returns the current state of the widget."""

        return self.state_machine()

    @property
    def styles(self) -> dict[str, BoundStyle]:
        """Returns a dictionary of current state's style keys to bound styles.

        Note that the `fill` style is inserted into every other style, and is
        stored under the `_fill` key for special circumstances where you may
        need to reference it.
        """

        styles = self.style_map[self.state.split("/")[0]].copy()

        if "/" in self.state:
            key = "/" + self.state.split("/")[-1]

            if key in self.style_map:
                styles |= self.style_map[key].items()

        output = {}
        fill = styles["fill"]

        if fill != "":
            fill += " "

        for key, style in styles.items():
            if key == "fill":
                key = "_fill"

            output[key] = BoundStyle(style, fill)

        return output

    @property
    def selected_index(self) -> int | None:
        """Returns the currently selected child's index."""

        return self._selected_index

    @property
    def selected(self) -> Widget | None:
        """Returns the currently selected widget."""

        if self.selected_index is None:
            return None

        return self

    @property
    def selectables(self) -> list[tuple[Widget, int]]:
        """Generates a list of tuples for selecting by the parent.

        See Container's selectables for more info.
        """

        return [(self, 0)]

    @property
    def selectable_count(self) -> int:
        """Returns the amount of selectable objects this widget has within."""

        return len(self.selectables)

    @property
    def frame(self) -> Frame:
        """Returns and sets the current frame."""

        return self._frame

    @frame.setter
    def frame(self, new: str | tuple[str, str, str, str] | Type[Frame]) -> None:
        """Sets the frame setting."""

        if isinstance(new, tuple):
            sides = tuple(get_frame(side) for side in new)
            self._frame = Frame.compose(sides)  # type: ignore
            return

        if isinstance(new, str) or new is None:
            new = get_frame(new)

        self._frame = new()

    @property
    def alignment(self) -> tuple[Alignment, Alignment]:
        """Returns and sets the current alignment settings.

        Args:
            new: A tuple of either two strings (start, center, end) or two alignment
                constants.
        """

        return self._alignment

    @alignment.setter
    def alignment(self, new: tuple[Alignment, Alignment] | tuple[str, str]) -> None:
        """Sets the alignment setting."""

        if not all(isinstance(item, Alignment) for item in new):
            new = Alignment(new[0]), Alignment(new[1])

        assert isinstance(new[0], Alignment) and isinstance(new[1], Alignment)
        self._alignment = new

    @property
    def overflow(self) -> tuple[Overflow, Overflow]:
        """Returns and sets the current overflow settings.

        Args:
            new: A tuple of either two strings (start, center, end) or two alignment
                constants.
        """

        return self._overflow

    @overflow.setter
    def overflow(self, new: tuple[Overflow, Overflow] | tuple[str, str]) -> None:
        """Sets the overflow setting."""

        if not all(isinstance(item, Overflow) for item in new):
            new = Overflow(new[0]), Overflow(new[1])

        assert isinstance(new[0], Overflow) and isinstance(new[1], Overflow)
        self._overflow = new

    @property
    def positioning(self) -> Positioning:
        """Returns and sets the positioning strategy."""

        return self._positioning

    @positioning.setter
    def positioning(self, new: str | Positioning) -> None:
        """Sets the positioning strategy."""

        if isinstance(new, str):
            new = Positioning(new)

        self._positioning = new

    @property
    def _framed_width(self) -> int:
        """Gets the widget's width excluding its frame."""

        return max(self.computed_width - self.frame.width, 0)

    @property
    def _framed_height(self) -> int:
        """Gets the widget's height excluding its frame."""

        return max(self.computed_height - self.frame.height, 0)

    def _parse_markup(self, markup: str) -> tuple[Span, ...]:
        """Parses some markup into a span of tuples.

        This also handles (ignores) zenith's FULL_RESET spans, and replaces `/`
        unsetters with `/ {fill_style} {content_style}`.
        """

        fill = self.styles["_fill"]

        if isinstance(self.parent, Widget) and fill.style == "":
            fill = self.parent.styles["_fill"]

        content_style = fill.style + " " + self.styles["content"].style

        # Replace full unsetters with full unsetter + content style
        markup = RE_FULL_UNSETTER.sub("/ " + content_style, markup)
        markup = zml_pre_process(markup)

        return tuple(span for span in zml_get_spans(markup) if span is not FULL_RESET)

    def _add_scrollbars(  # pylint: disable=too-many-arguments
        self,
        lines: list[tuple[Span, ...]],
        width: int,
        height: int,
        scrollbar_x: bool,
        scrollbar_y: bool,
    ) -> None:
        """Adds scrollbars to the given lines."""

        if scrollbar_x:
            buff = "".join(
                list(
                    _build_scrollbar(
                        *self.frame.scrollbars[0],
                        size=width - scrollbar_y,
                        current=self.scroll[0] / max(self._virtual_width - width, 1),
                        ratio=width / self._virtual_width,
                    )
                )
            )

            lines[-1] = self._parse_markup(
                self.styles["scrollbar_x"](buff + " " * scrollbar_y)
            )

        if scrollbar_y:
            chars = list(
                _build_scrollbar(
                    *self.frame.scrollbars[1],
                    size=height - scrollbar_x,
                    current=self.scroll[1] / max(self._virtual_height - height, 1),
                    ratio=height / self._virtual_height,
                )
            )

            for i, line in enumerate(lines):
                if i >= height - scrollbar_x:
                    break

                span = line[-1]

                lines[i] = (
                    *line[:-1],
                    span.mutate(text=span.text[:-1]),
                    *(self._parse_markup(self.styles["scrollbar_y"](chars[i]))),
                )

    def _apply_frame(  # pylint: disable=too-many-locals
        self, lines: list[tuple[Span, ...]], width: int
    ) -> None:
        """Adds frame characters around the given lines."""

        def _style(item, outer: bool = False) -> tuple[Span, ...]:
            style = deepcopy(self.styles["frame"])

            if outer:
                if self.parent is not None and hasattr(self.parent, "styles"):
                    style.style = self.parent.styles["frame"].fill + " " + style.style

            return self._parse_markup(style(item))

        left_top, right_top, right_bottom, left_bottom = self.frame.corners
        left, top, right, bottom = self.frame.borders

        h_outer = self.frame.outer_horizontal
        v_outer = self.frame.outer_vertical
        c_outer = self.frame.outer_corner

        for i, line in enumerate(lines):
            lines[i] = (
                *_style(left, outer=v_outer),
                *line,
                *_style(right, outer=v_outer),
            )

        if left_top + top + right_top != "":
            lines.insert(
                0,
                (
                    *_style(left_top, outer=c_outer),
                    *_style(top * width, outer=h_outer),
                    *_style(right_top, outer=c_outer),
                ),
            )

        if left_bottom + bottom + right_bottom != "":
            lines.append(
                (
                    *_style(left_bottom, outer=c_outer),
                    *_style(bottom * width, outer=h_outer),
                    *_style(right_bottom, outer=c_outer),
                )
            )

    def _horizontal_align(self, line: tuple[Span, ...], width: int) -> tuple[Span, ...]:
        """Aligns a tuple of spans horizontally, using `self.alignment[0]`."""

        width = max(width, self._virtual_width)
        length = sum(len(span.text) for span in line)
        diff = width - length

        style = self.styles["content"]

        if line in [tuple(), (Span(""),)]:
            return self._parse_markup(style(diff * " "))

        alignment = self.alignment[0]

        if alignment is Alignment.START:
            return (*line[:-1], line[-1].mutate(text=line[-1].text + diff * " "))

        if alignment is Alignment.CENTER:
            end, extra = divmod(diff, 2)
            start = end + extra

            # Return padding as a single span instead of (start, content, end)
            # when content is " " (filler lines)
            if "".join(span.text for span in line) == " ":
                return self._parse_markup(style((start + end) * " "))

            return (
                *self._parse_markup(style(start * " ")),
                line[0],
                *line[1:],
                *self._parse_markup(style(end * " ")),
            )

        if alignment is Alignment.END:
            span = line[0]

            return (span.mutate(text=diff * " " + span.text), *line[1:])

        raise NotImplementedError(f"Unknown alignment {alignment!r}.")

    def _vertical_align(self, lines: list[tuple[Span, ...]], height: int) -> None:
        """Aligns a list of tuples of spans vertically, using `self.alignment[1]`.

        Note that this mutates `lines`.
        """

        available = height - len(lines)
        alignment = self.alignment[1]
        filler = self._parse_markup(self.styles["content"](" "))

        if alignment is Alignment.START:
            lines.extend([filler] * available)
            return

        if alignment is Alignment.CENTER:
            top, extra = divmod(available, 2)
            bottom = top + extra

            for _ in range(top):
                lines.insert(0, filler)

            lines.extend([filler] * bottom)
            return

        for _ in range(available):
            lines.insert(0, filler)

    def _slice_line(
        self, line: tuple[Span, ...], start: int, end: int
    ) -> tuple[Span, ...]:
        """Slices a line into the given width."""

        if end is None:
            end = len(line) - 1

        width = end - start

        line_list = []
        before_start = 0
        occupied = 0

        for span in line:
            length = len(span)

            before_start += length
            if before_start < start:
                continue

            new = span[max(start - (before_start - length), 0) :]
            length = len(new)

            line_list.append(new)

            occupied += length

            if occupied > width:
                break

        width_diff = max(occupied - width, 0)

        if width_diff > 0 and len(line_list) > 0:
            for i, span in enumerate(reversed(line_list)):
                new = span[:-width_diff]
                width_diff -= len(span) - len(new)

                line_list[-i - 1] = new

                if width_diff <= 0:
                    break

        if not line_list:
            line_list.extend(self._parse_markup(self.styles["content"](" ")))

        occupied = sum(len(span) for span in line_list)

        if occupied < width:
            suffix = line_list[-1]

            line_list[-1] = suffix.mutate(text=suffix.text + (width - occupied) * " ")

        return tuple(line_list)

    def _apply_mouse_state(self, action: MouseAction) -> None:
        """Applies a state change action based on mouse input."""

        value = action.value

        if "click" in value:
            self.state_machine.apply_action("CLICKED")
            self.state_machine.apply_action("SUBSTATE_EXIT_SCROLLING_X")
            self.state_machine.apply_action("SUBSTATE_EXIT_SCROLLING_Y")
            return

        if "release" in value:
            self.state_machine.apply_action("RELEASED")
            self.state_machine.apply_action("SUBSTATE_EXIT_SCROLLING_X")
            self.state_machine.apply_action("SUBSTATE_EXIT_SCROLLING_Y")

            if self.selected_index is not None:
                self.state_machine.apply_action("SELECTED")

            return

        if "hover" in value:
            self.state_machine.apply_action("HOVERED")
            return

    def setup(self) -> None:
        """Use this to do simple setup actions without overriding __init__."""

    def serialize(self) -> dict[str, Any]:
        """Returns a dictionary to represent the dynamic value of the widget."""

        return {}

    def update(
        self,
        attrs: dict[str, Any],
        style_map: dict[str, str],
    ) -> None:
        """Updates a widget with new setttings.

        Args:
            attrs: A dictionary of the widget's fields to values. Fields starting with
                `_` are disallowed, as are fields the widget doesn't already have.
            style_map: A dictionary of style keys to markup that gets merged onto our
                `style_map`'s current state, i.e.:

                    `self.style_map | {self.state: style_map}`
        """

        keys = dir(self)

        for key, value in attrs.items():
            if key.startswith("_"):
                raise ValueError(f"cannot set non-public attr {key!r}")

            if key not in keys:
                raise ValueError(f"cannot set non-existant attr {key!r}")

            setattr(self, key, value)

        self.style_map = self.style_map | {self.state: style_map}

    def as_query(self, state: bool = False) -> str:
        """Returns the widget as the most specific selectable query.

        Args:
            state: Include state suffix onto the query.
        """

        query = type(self).__name__

        if self.eid is not None:
            query += f"#{self.eid}"

        for variant in self.groups:
            query += f".{variant}"

        if state:
            query += f"/{self.state}"

        return query

    def query_changed(self) -> bool:
        """Returns whether the result of `as_query` has changed since last call."""

        query = self.as_query(state=True)
        value = query != self._last_query

        self._last_query = query
        return value

    def is_static(self) -> bool:
        """Shorthand for `positioning is Positioning.STATIC`."""

        return self.positioning is Positioning.STATIC

    def is_dynamic(self) -> bool:
        """Shorthand for `positioning is Positioning.DYNAMIC`."""

        return self.positioning is Positioning.DYNAMIC

    def is_fill_width(self) -> bool:
        """Shorthand for `width is None`."""

        return self.width is None

    def is_fill_height(self) -> bool:
        """Shorthand for `height is None`."""

        return self.height is None

    def scroll_to(self, x: int | None = None, y: int | None = None) -> None:
        """Scrolls the widget.

        Both arguments may be given in one of four forms:

        - `0`: scroll to the start of the widget on the given axis
        - `-1`: scroll to the end of the widget on the given axis
        - `0<` : scroll to the offset on the given axis
        - `None`: do nothing for the given axis.

        Args:
            x: The x scroll value.
            y: The y scroll value.
        """

        new_x = x if x is not None else self.scroll[0]

        if new_x == -1:
            new_x = self._virtual_width

        new_y = y if y is not None else self.scroll[1]

        if new_y == -1:
            new_y = self._virtual_height

        self.scroll = (new_x, new_y)

    def has_scrollbar(self, index: Literal[0, 1]) -> bool:
        """Returns whether the given dimension should display a scrollbar.

        Args:
            index: The axis to check. 0 for horizontal scrolling, 1 for vertical.
        """

        real, virt = [
            (self._framed_width, self._virtual_width),
            (self._framed_height, self._virtual_height),
        ][index]

        return self.overflow[index] is Overflow.SCROLL or (
            self.overflow[index] is Overflow.AUTO and _overflows(real, virt)
        )

    def drawables(self) -> Iterable[Widget]:
        """Yields all contained widgets that should be drawn."""

        yield self

    def clip_height(self, start: int | None, end: int | None) -> None:
        """Sets the clipping rectangle's coordinates."""

        self._clip_start = start
        self._clip_end = end

    def add_group(self, target: str) -> None:
        """Adds a group to the widget's groups."""

        self.groups = tuple(list(self.groups) + [target])

    def remove_group(self, target: str) -> None:
        """Removes a group from the widget's groups."""

        self.groups = tuple(group for group in self.groups if group != target)

    def toggle_group(self, target: str) -> bool:
        """Toggles a group in the widget's groups."""

        if target in self.groups:
            self.remove_group(target)
            return False

        self.add_group(target)
        return True

    def select(self, index: int | None = None) -> None:
        """Selects a part of this Widget.

        Args:
            index: The index to select.

        Raises:
            TypeError: This widget has no selectables, i.e. widget.is_selectable == False.
        """

        if index is not None:
            index = min(max(0, index), self.selectable_count - 1)

        self._selected_index = index
        self.state_machine.apply_action(("UN" if index is None else "") + "SELECTED")

    def hide(self) -> None:
        """Adds the `hidden` group onto the widget."""

        self.add_group("hidden")

    def show(self) -> None:
        """Remove the `hidden` group from the widget."""

        self.remove_group("hidden")

    def contains(self, position: tuple[int, int]) -> bool:
        """Determines whether this widget contains the given position."""

        rect = self.position, (
            self.position[0] + self.computed_width - self.has_scrollbar(1),
            self.position[1] + self.computed_height - self.has_scrollbar(0),
        )

        (left, top), (right, bottom) = rect

        return left < position[0] <= right and top < position[1] <= bottom

    def move_to(self, x: int, y: int) -> None:
        """Moves the widget to the given position."""

        self.position = x, y

    def move_by(self, x: int, y: int) -> None:
        """Moves the widget by the given changes."""

        self.move_to(self.position[0] + x, self.position[1] + y)

    def bind(self, key: str, callback: Callable[[], Any]) -> Event:
        """Binds a key to the given callback.

        If a binding already exists, the given callback is added to the Event.
        """

        if key not in self._bindings:
            self._bindings[key] = Event(f"on key {key}")

        event = self._bindings[key]
        event += callback

        return event

    def unbind(self, key: str) -> Event:
        """Removes all bindings for a given key."""

        if key not in self._bindings:
            raise ValueError(f"can't remove non-existant binding for key {key!r}.")

        event = self._bindings[key]
        del self._bindings[key]

        return event

    def handle_keyboard(self, key: Key) -> bool:
        """Handles a keyboard event.

        Returns:
            Whether the event should continue to bubble upwards. Return `True` if the
            event was already handled.
        """

        result = False

        for value in key:
            if value in self._bindings:
                if self._bindings[value](self):
                    result |= True

            if result:
                break

        return result

    def handle_mouse(self, action: MouseAction, position: tuple[int, int]) -> bool:
        """Handles a mouse event.

        Returns:
            Whether the event should continue to bubble upwards. Return `True` if the
            event was already handled.
        """

        self._apply_mouse_state(action)

        if "scroll" in action.value:
            can_scroll_x, can_scroll_y = self.has_scrollbar(0), self.has_scrollbar(1)
            if (
                can_scroll_x
                and action is MouseAction.SCROLL_LEFT
                or action is MouseAction.SHIFT_SCROLL_UP
            ):
                self.scroll = (self.scroll[0] - self.scroll_step, self.scroll[1])

            elif (
                can_scroll_x
                and action is MouseAction.SCROLL_RIGHT
                or action is MouseAction.SHIFT_SCROLL_DOWN
            ):
                self.scroll = (self.scroll[0] + self.scroll_step, self.scroll[1])

            elif can_scroll_y and action is MouseAction.SCROLL_UP:
                self.scroll = (self.scroll[0], self.scroll[1] - self.scroll_step)

            elif can_scroll_y and action is MouseAction.SCROLL_DOWN:
                self.scroll = (self.scroll[0], self.scroll[1] + self.scroll_step)

        def _get_names(action: MouseAction) -> tuple[str, ...]:
            if action.value == "hover":
                return (action.value,)

            parts = action.value.split("_")

            full = action.value
            alt_key = "_".join(parts[-2:])
            event = parts[-1]

            if alt_key in ["left", "right"]:
                return (full, alt_key, event)

            return (full, event)

        possible_names = _get_names(action)

        for name in possible_names:
            if hasattr(self, f"on_{name}"):
                handle = getattr(self, f"on_{name}")

                return handle(action, position)

        return False

    def compute_dimensions(self, available_width: int, available_height: int) -> None:
        """Computes width & height based on our specifications and the given space."""

        self.computed_width = _compute(self.width, available_width)
        self.computed_height = _compute(self.height, available_height)

    def get_content(self) -> list[str]:
        """Gets the dynamic content for this widget."""

        raise NotImplementedError

    def build(
        self, *, virt_width: int | None = None, virt_height: int | None = None
    ) -> list[tuple[Span, ...]]:
        """Builds the strings that represent the widget."""

        width = self._framed_width
        height = self._framed_height

        def _clamp_scrolls() -> tuple[int, int]:
            x_bar = _overflows(width, self._virtual_width)
            y_bar = _overflows(height, self._virtual_height)

            return (
                max(min(self.scroll[0], self._virtual_width - width + x_bar), 0),
                max(min(self.scroll[1], self._virtual_height - height + y_bar), 0),
            )

        self.scroll = _clamp_scrolls()

        content_style = self.styles["content"]

        lines: list[tuple[Span, ...]] = [
            self._parse_markup(content_style(line)) for line in self.get_content()
        ]

        self._virtual_height = virt_height or len(lines)
        self._virtual_width = virt_width or max(
            sum(len(span) for span in line) for line in lines
        )

        # Clip into vertical size
        if self._virtual_height > height:
            lines = lines[self.scroll[1] : self.scroll[1] + height]

            if len(lines) < height:
                lines.extend([(Span(""),)] * (height - len(lines)))

        # Handle alignment
        self._vertical_align(lines, height)

        for i, line in enumerate(lines):
            lines[i] = self._horizontal_align(line, width)

        # Clip into horizontal size
        lines = [
            self._slice_line(line, self.scroll[0], self.scroll[0] + width)
            for line in lines
        ]

        # Composite together frame + content + scrollbar
        self._add_scrollbars(
            lines,
            width,
            height,
            self.has_scrollbar(0),
            self.has_scrollbar(1),
        )

        self._apply_frame(lines, width)

        if len(lines) > self.computed_height:
            lines = lines[: self.computed_height]

        lines = lines[self._clip_start : self._clip_end]

        return lines


widget_annotations = Widget.__annotations__
widget_types: dict[str, Type[Widget]] = {"Widget": Widget}


def handle_mouse_on_children(
    action: MouseAction,
    position: tuple[int, int],
    mouse_target: Widget | None,
    hover_target: Widget | None,
    children: Iterable[Widget],
) -> tuple[bool, int | None, Widget | None, Widget | None]:
    """Handles the given mouse event on an iterable of widgets.

    This can be used by 'container' type widgets (like `Container`), or meta-level
    management objects (like `Application`) for consistent mouse handling behaviour.

    Args:
        action: The action associated with the event.
        position: The position associated with the event.
        mouse_target: The caller's current mouse target.
        hover_target: The caller's current hover target.
        children: The widgets to execute upon.

    Returns:
        A tuple of:

            (success, selection_index, mouse_target, hover_target)

        Since this function doesn't have access to the caller, the caller has to set
        its mouse & hover targets to the returned values. `selection_index` is only set
        when a new widget handled the event & is selectable, so should be ignored when
        set to None.
    """

    if action is MouseAction.LEFT_RELEASE:
        if mouse_target is not None:
            mouse_target.handle_mouse(action, position)
            mouse_target = None

        if hover_target is not None:
            hover_target.handle_mouse(action, position)
            hover_target = None

        return True, None, mouse_target, hover_target

    is_hover = action is MouseAction.HOVER
    release = MouseAction.LEFT_RELEASE, position

    if (
        mouse_target is not None
        and "click" not in action.value  # Clicks cannot be done outside of the widget
        and mouse_target.handle_mouse(action, position)
    ):
        return True, None, mouse_target, hover_target

    if is_hover and hover_target is not None and not hover_target.contains(position):
        hover_target.handle_mouse(*release)
        hover_target = None

    selection = 0

    for child in children:
        selection += child.selectable_count

        if not child.contains(position):
            continue

        if is_hover:
            if hover_target is not None and hover_target is not child:
                hover_target.handle_mouse(*release)

            hover_target = child

        if child.handle_mouse(action, position):
            selection -= child.selectable_count - (child.selected_index or 0)

            if mouse_target is not None and mouse_target is not child:
                mouse_target.handle_mouse(*release)

            mouse_target = child

            return True, selection, mouse_target, hover_target

        break

    return False, None, mouse_target, hover_target



from __future__ import annotations

from typing import Any

from slate import Event, EventCallback

from ..enums import MouseAction
from .button import Button
from .widget import Widget


class Checkbox(Widget):
    """A button-like widget for checkboxes."""

    content: str
    checked: bool = False
    indicators: tuple[str, str] = ("□", "▣")

    on_change: Event
    """Called when the Checkbox's value changed.

    Args:
        self: The Checkbox that sent the event. Use `self.checked` to get current state.
    """

    style_map = Button.style_map | {
        state: {"indicator": ""} for state in Button.state_machine.states
    }

    # TODO: Add a form of inheritance to YAML styling
    rules = """
    Checkbox:
        content_style: dim

        height: 1

        /idle:
            frame_style: ui.primary+1

        /hover:
            fill_style: '@ui.panel1-2'
            frame_style: ui.primary+1

        /selected:
            content_style: dim bold
            fill_style: '@ui.panel1-1'
            frame_style: ui.primary+1

        /active:
            fill_style: '@ui.primary+3'
            frame_style: 'ui.panel1'
            indicator_style: 'dim'
    """

    def __init__(
        self,
        content: str,
        checked: bool = False,
        name: str | None = None,
        on_change: list[EventCallback] | None = None,
        **widget_args: Any,
    ) -> None:
        super().__init__(**widget_args)

        self._has_timeout = False

        self.content = content
        self.checked = checked
        self.name = name

        self.on_change = Event("checkbox changed")

        for event in on_change or []:
            self.on_change += event

        self.bind("return", self._visual_submit)

    def _visual_submit(self) -> bool:
        """Animates the active state when the button is submitted using a keyboard."""

        if self._has_timeout:
            return False

        self.checked = not self.checked
        self.state_machine.apply_action("CLICKED")

        self._has_timeout = True
        self.app.timeout(
            150,
            lambda: (
                self.state_machine.apply_action("RELEASED_KEYBOARD"),
                setattr(self, "_has_timeout", False),  # type: ignore
            ),
        )

        return self.on_change(self)

    def on_click(self, _: MouseAction, __: tuple[int, int]) -> bool:
        """Toggles checked & emits the change event."""

        self.checked = not self.checked
        self.on_change(self.checked)

        return True

    def serialize(self) -> dict[str, Any]:
        if self.name is None:
            raise ValueError(f"field {self!r} cannot be serialized without a name.")

        return {self.name: self.checked}

    def get_content(self) -> list[str]:
        indicator = self.styles["indicator"](self.indicators[self.checked]) + "[/]"

        return [f" {indicator} {self.content} "]



from slate import Color
from zenith import Palette

palette = Palette(Color.from_hex("#AFE1AF").darken(2))

palette.color_mapping = {
    "ui.text": Color((245, 245, 245)).darken(2),
    "ui.primary": palette.primary,
    "ui.secondary": palette.secondary,
    "ui.accent": palette.quaternary,
    "ui.panel1": palette.surface1.darken(5),
    "ui.panel2": palette.surface2.darken(5),
    "ui.panel3": palette.surface3.darken(5),
    "ui.panel4": palette.surface4.darken(5),
    "ui.success": palette.success,
    "ui.warning": palette.warning,
    "ui.error": palette.error,
}

palette.alias()



from __future__ import annotations

import re
from typing import Any

from .widget import Widget
from ..enums import MouseAction

RE_HYPERLINK_MARKUP = re.compile(r"~([^ \]\[]+)")


class Text(Widget):
    """A widget that displays some static text."""

    rules = """
    Text:
        height: 1
    """

    def __init__(self, content: str, **widget_args: Any) -> None:
        super().__init__(**widget_args)

        self.content = content

        self.width = max(
            (len(span) for span in self._parse_markup(content.splitlines()[0])),
            default=0,
        )

    @classmethod
    def lorem(cls, **widget_args: Any) -> Text:
        """Returns a Text object filled with Lorem Ipsum."""

        return Text(
            **widget_args,
            groups=("fill-height", "overflow-auto"),
            content="""\
            Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum in metus
            suscipit, luctus augue vel, fringilla erat. Nulla at turpis volutpat,
            porttitor lectus eu, malesuada eros. Phasellus at tortor libero. Class
            aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos
            himenaeos. Nam pellentesque dolor a scelerisque feugiat. Duis nec ante mi.
            Integer in tortor elementum, imperdiet nunc sed, tincidunt erat. Sed non mi
            accumsan, aliquet ante convallis, vulputate velit. Nullam malesuada nisl
            ligula, ut fermentum risus tincidunt at. Vivamus sollicitudin condimentum
            maximus. Etiam eu metus rhoncus, placerat orci ut, rutrum lorem.

            Aliquam at laoreet ipsum. In vitae ante eu eros consectetur porta.
            Curabitur euismod, massa in tempor convallis, neque quam elementum sem, eu
            ultrices ex odio sed tellus. Curabitur ac ligula dignissim, lobortis lectus
            vel, condimentum risus. Praesent quis accumsan tortor, vitae aliquam neque.
            Maecenas id fermentum ex. Etiam enim purus, viverra et elementum vitae,
            egestas et libero. Etiam vel lacus imperdiet, viverra purus quis, posuere
            justo. Duis eu ex suscipit nisi hendrerit varius auctor id metus. Nullam
            quis sollicitudin nibh. Nam eleifend pellentesque sodales. Vestibulum
            ullamcorper vulputate euismod.

            Curabitur ultrices commodo metus, suscipit varius velit. Nam sem tortor,
            volutpat id sagittis at, cursus in purus. Fusce pretium, tellus non ultrices
            feugiat, mauris purus condimentum lacus, id vestibulum eros velit vel orci.
            Praesent gravida turpis et feugiat placerat. Sed at tellus congue, semper
            diam aliquam, elementum ipsum. Nam eu libero lobortis, consectetur erat ege,
            viverra est. Donec ultricies velit at dui egestas ultricies. Suspendisse
            gravida velit non interdum sodales. Donec finibus ac velit non facilisis.
            Praesent ipsum neque, malesuada ut laoreet sit amet, eleifend aliquet quam.
            Fusce ex nibh, fringilla a augue eget, tristique lacinia ligula. Ut eu
            tortor non dolor ultrices hendrerit eget non lectus. Mauris pharetra dolor
            sit amet ante pellentesque sagittis. Praesent in ultricies odio, eu laoreet
            felis.

            In a augue sollicitudin felis commodo sollicitudin nec vel nibh. Nunc
            euismod semper massa, in finibus enim lacinia consectetur. Ut tristique
            risus augue, in mollis elit sodales at. Donec sed ipsum ligula. Praesent
            suscipit neque ut faucibus euismod. Duis pellentesque ac enim eu ornare.
            Curabitur blandit, velit ut luctus semper, erat sapien fermentum velit, a
            sodales leo arcu nec est.

            Morbi rutrum in est vitae placerat. Orci varius natoque penatibus et magnis
            dis parturient montes, nascetur ridiculus mus. Nullam congue non mi at
            malesuada. Nulla velit dui, iaculis in viverra id, facilisis at nibh. In
            quis metus a erat lacinia porttitor sed vitae elit. Vivamus scelerisque
            sagittis ex, quis scelerisque leo consectetur eget. Nunc sed sodales mi.
            Nunc viverra dui eget erat accumsan, id placerat ligula consectetur. Sed
            elementum dictum tortor ultrices aliquet. Proin ullamcorper euismod congue.
            Nullam eleifend lacus euismod ante mattis sagittis. Phasellus eu erat mauri.

            Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum in metus
            suscipit, luctus augue vel, fringilla erat. Nulla at turpis volutpat,
            porttitor lectus eu, malesuada eros. Phasellus at tortor libero. Class
            aptent taciti sociosqu ad litora torquent per conubia nostra, per inceptos
            himenaeos. Nam pellentesque dolor a scelerisque feugiat. Duis nec ante mi.
            Integer in tortor elementum, imperdiet nunc sed, tincidunt erat. Sed non mi
            accumsan, aliquet ante convallis, vulputate velit. Nullam malesuada nisl
            ligula, ut fermentum risus tincidunt at. Vivamus sollicitudin condimentum
            maximus. Etiam eu metus rhoncus, placerat orci ut, rutrum lorem.

            Aliquam at laoreet ipsum. In vitae ante eu eros consectetur porta.
            Curabitur euismod, massa in tempor convallis, neque quam elementum sem, eu
            ultrices ex odio sed tellus. Curabitur ac ligula dignissim, lobortis lectus
            vel, condimentum risus. Praesent quis accumsan tortor, vitae aliquam neque.
            Maecenas id fermentum ex. Etiam enim purus, viverra et elementum vitae,
            egestas et libero. Etiam vel lacus imperdiet, viverra purus quis, posuere
            justo. Duis eu ex suscipit nisi hendrerit varius auctor id metus. Nullam
            quis sollicitudin nibh. Nam eleifend pellentesque sodales. Vestibulum
            ullamcorper vulputate euismod.

            Curabitur ultrices commodo metus, suscipit varius velit. Nam sem tortor,
            volutpat id sagittis at, cursus in purus. Fusce pretium, tellus non ultrices
            feugiat, mauris purus condimentum lacus, id vestibulum eros velit vel orci.
            Praesent gravida turpis et feugiat placerat. Sed at tellus congue, semper
            diam aliquam, elementum ipsum. Nam eu libero lobortis, consectetur erat ege,
            viverra est. Donec ultricies velit at dui egestas ultricies. Suspendisse
            gravida velit non interdum sodales. Donec finibus ac velit non facilisis.
            Praesent ipsum neque, malesuada ut laoreet sit amet, eleifend aliquet quam.
            Fusce ex nibh, fringilla a augue eget, tristique lacinia ligula. Ut eu
            tortor non dolor ultrices hendrerit eget non lectus. Mauris pharetra dolor
            sit amet ante pellentesque sagittis. Praesent in ultricies odio, eu laoreet
            felis.

            In a augue sollicitudin felis commodo sollicitudin nec vel nibh. Nunc
            euismod semper massa, in finibus enim lacinia consectetur. Ut tristique
            risus augue, in mollis elit sodales at. Donec sed ipsum ligula. Praesent
            suscipit neque ut faucibus euismod. Duis pellentesque ac enim eu ornare.
            Curabitur blandit, velit ut luctus semper, erat sapien fermentum velit, a
            sodales leo arcu nec est.

            Morbi rutrum in est vitae placerat. Orci varius natoque penatibus et magnis
            dis parturient montes, nascetur ridiculus mus. Nullam congue non mi at
            malesuada. Nulla velit dui, iaculis in viverra id, facilisis at nibh. In
            quis metus a erat lacinia porttitor sed vitae elit. Vivamus scelerisque
            sagittis ex, quis scelerisque leo consectetur eget. Nunc sed sodales mi.
            Nunc viverra dui eget erat accumsan, id placerat ligula consectetur. Sed
            elementum dictum tortor ultrices aliquet. Proin ullamcorper euismod congue.
            Nullam eleifend lacus euismod ante mattis sagittis. Phasellus eu erat mauri.
            """,
        )

    @property
    def selectables(self) -> list[tuple[Widget, int]]:
        return []

    def on_click(self, _: MouseAction, __: tuple[int, int]) -> bool:
        """Clicks the first hyperlink within the text's content."""

        if self.app is None:
            return False

        if (mtch := RE_HYPERLINK_MARKUP.search(self.content)) is not None:
            value = mtch[1]

            self.app.route(value)

        # Never return True, as we don't want the parent to select a widget
        return False

    def get_content(self) -> list[str]:
        return self.content.splitlines()

    # Proper linebreaking prototype
    #
    # lines = []
    # usable_width = self.width - self.frame.width - 1
    #
    # for line in self.content.splitlines():
    #     length = 0
    #     buff = ""
    #
    #     for char in line:
    #         if char == "[":
    #             in_tag = True
    #
    #         elif char == "]":
    #             in_tag = False
    #
    #         if length < usable_width and char not in ["\n", "\r"]:
    #             if char != "]" and not in_tag:
    #                 length += 1
    #
    #             buff += char
    #             continue
    #
    #         lines.append(buff)
    #         length = 0
    #         buff = ""
    #
    #     lines.append(buff)
    #
    # return lines



from enum import Enum

__all__ = [
    "Alignment",
    "Overflow",
    "MouseAction",
]


class Alignment(Enum):
    """An enum used to select alignment strategy."""

    START = "start"
    """Align to the start of the area (left or top)."""

    CENTER = "center"
    """Align to centered in the area."""

    END = "end"
    """Align to the end of the area (right or bottom)."""


class Overflow(Enum):
    """An enum used to select overflow strategy."""

    HIDE = "hide"
    """Hide scrollbar for overflowing content."""

    SCROLL = "scroll"
    """Display scrollbar for overflowing content."""

    AUTO = "auto"
    """Display scrollbar only when there is overflowing content."""


class Direction(Enum):
    """An enum used for selecting layout direction."""

    HORIZONTAL = "horizontal"
    """Flow content horizontally."""

    VERTICAL = "vertical"
    """Flow content vertically."""


class Positioning(Enum):
    """An enum used to represent a positioning strategy."""

    DYNAMIC = "dynamic"
    """Widget should be arranged by its parent."""

    STATIC = "static"
    """Widget should be excluded from parent's arrangement, except for sizing."""


class MouseAction(Enum):
    """A humonculous enumeration of mouse actions."""

    LEFT_CLICK = "left_click"
    LEFT_RELEASE = "left_release"
    RIGHT_CLICK = "right_click"
    RIGHT_RELEASE = "right_release"
    SHIFT_LEFT_CLICK = "shift_left_click"
    SHIFT_LEFT_RELEASE = "shift_left_release"
    SHIFT_RIGHT_CLICK = "shift_right_click"
    SHIFT_RIGHT_RELEASE = "shift_right_release"
    OPTION_LEFT_CLICK = "option_left_click"
    OPTION_LEFT_RELEASE = "option_left_release"
    OPTION_RIGHT_CLICK = "option_right_click"
    OPTION_RIGHT_RELEASE = "option_right_release"
    SHIFT_OPTION_LEFT_CLICK = "shift_option_left_click"
    SHIFT_OPTION_LEFT_RELEASE = "shift_option_left_release"
    SHIFT_OPTION_RIGHT_CLICK = "shift_option_right_click"
    SHIFT_OPTION_RIGHT_RELEASE = "shift_option_right_release"
    CTRL_LEFT_CLICK = "ctrl_left_click"
    CTRL_LEFT_RELEASE = "ctrl_left_release"
    CTRL_RIGHT_CLICK = "ctrl_right_click"
    CTRL_RIGHT_RELEASE = "ctrl_right_release"
    SHIFT_CTRL_LEFT_CLICK = "shift_ctrl_left_click"
    SHIFT_CTRL_LEFT_RELEASE = "shift_ctrl_left_release"
    SHIFT_CTRL_RIGHT_CLICK = "shift_ctrl_right_click"
    SHIFT_CTRL_RIGHT_RELEASE = "shift_ctrl_right_release"
    CTRL_OPTION_LEFT_CLICK = "ctrl_option_left_click"
    CTRL_OPTION_LEFT_RELEASE = "ctrl_option_left_release"
    CTRL_OPTION_RIGHT_CLICK = "ctrl_option_right_click"
    CTRL_OPTION_RIGHT_RELEASE = "ctrl_option_right_release"
    SHIFT_CTRL_OPTION_LEFT_CLICK = "shift_ctrl_option_left_click"
    SHIFT_CTRL_OPTION_LEFT_RELEASE = "shift_ctrl_option_left_release"
    SHIFT_CTRL_OPTION_RIGHT_CLICK = "shift_ctrl_option_right_click"
    SHIFT_CTRL_OPTION_RIGHT_RELEASE = "shift_ctrl_option_right_release"
    LEFT_DRAG = "left_drag"
    RIGHT_DRAG = "right_drag"
    HOVER = "hover"
    SHIFT_LEFT_DRAG = "shift_left_drag"
    SHIFT_RIGHT_DRAG = "shift_right_drag"
    SHIFT_HOVER = "shift_hover"
    OPTION_LEFT_DRAG = "option_left_drag"
    OPTION_RIGHT_DRAG = "option_right_drag"
    OPTION_HOVER = "option_hover"
    SHIFT_OPTION_LEFT_DRAG = "shift_option_left_drag"
    SHIFT_OPTION_RIGHT_DRAG = "shift_option_right_drag"
    SHIFT_OPTION_HOVER = "shift_option_hover"
    CTRL_LEFT_DRAG = "ctrl_left_drag"
    CTRL_RIGHT_DRAG = "ctrl_right_drag"
    CTRL_HOVER = "ctrl_hover"
    SHIFT_CTRL_LEFT_DRAG = "shift_ctrl_left_drag"
    SHIFT_CTRL_RIGHT_DRAG = "shift_ctrl_right_drag"
    SHIFT_CTRL_HOVER = "shift_ctrl_hover"
    CTRL_OPTION_LEFT_DRAG = "ctrl_option_left_drag"
    CTRL_OPTION_RIGHT_DRAG = "ctrl_option_right_drag"
    CTRL_OPTION_HOVER = "ctrl_option_hover"
    SHIFT_CTRL_OPTION_LEFT_DRAG = "shift_ctrl_option_left_drag"
    SHIFT_CTRL_OPTION_RIGHT_DRAG = "shift_ctrl_option_right_drag"
    SHIFT_CTRL_OPTION_HOVER = "shift_ctrl_option_hover"
    SCROLL_UP = "scroll_up"
    SCROLL_DOWN = "scroll_down"
    SCROLL_LEFT = "scroll_left"
    SCROLL_RIGHT = "scroll_right"
    SHIFT_SCROLL_UP = "shift_scroll_up"
    SHIFT_SCROLL_DOWN = "shift_scroll_down"
    SHIFT_SCROLL_LEFT = "shift_scroll_left"
    SHIFT_SCROLL_RIGHT = "shift_scroll_right"
    OPTION_SCROLL_UP = "option_scroll_up"
    OPTION_SCROLL_DOWN = "option_scroll_down"
    OPTION_SCROLL_LEFT = "option_scroll_left"
    OPTION_SCROLL_RIGHT = "option_scroll_right"
    SHIFT_OPTION_SCROLL_UP = "shift_option_scroll_up"
    SHIFT_OPTION_SCROLL_DOWN = "shift_option_scroll_down"
    SHIFT_OPTION_SCROLL_LEFT = "shift_option_scroll_left"
    SHIFT_OPTION_SCROLL_RIGHT = "shift_option_scroll_right"
    CTRL_SCROLL_UP = "ctrl_scroll_up"
    CTRL_SCROLL_DOWN = "ctrl_scroll_down"
    CTRL_SCROLL_LEFT = "ctrl_scroll_left"
    CTRL_SCROLL_RIGHT = "ctrl_scroll_right"
    SHIFT_CTRL_SCROLL_UP = "shift_ctrl_scroll_up"
    SHIFT_CTRL_SCROLL_DOWN = "shift_ctrl_scroll_down"
    SHIFT_CTRL_SCROLL_LEFT = "shift_ctrl_scroll_left"
    SHIFT_CTRL_SCROLL_RIGHT = "shift_ctrl_scroll_right"
    CTRL_OPTION_SCROLL_UP = "ctrl_option_scroll_up"
    CTRL_OPTION_SCROLL_DOWN = "ctrl_option_scroll_down"
    CTRL_OPTION_SCROLL_LEFT = "ctrl_option_scroll_left"
    CTRL_OPTION_SCROLL_RIGHT = "ctrl_option_scroll_right"
    SHIFT_CTRL_OPTION_SCROLL_UP = "shift_ctrl_option_scroll_up"
    SHIFT_CTRL_OPTION_SCROLL_DOWN = "shift_ctrl_option_scroll_down"
    SHIFT_CTRL_OPTION_SCROLL_LEFT = "shift_ctrl_option_scroll_left"
    SHIFT_CTRL_OPTION_SCROLL_RIGHT = "shift_ctrl_option_scroll_right"



from __future__ import annotations

from typing import Any

from slate import Event, EventCallback

from ..enums import MouseAction
from .widget import Widget

SEMANTIC_STYLES = """
        .{base}:
            /idle|hover:
                frame_style: ui.{base}

            /selected:
                frame_style: ui.{base}

            /active:
                frame_style: ui.panel1-2
                fill_style: '@ui.{base}+3'
"""


class Button(Widget):
    """A pressable/clickable button."""

    on_submit: Event
    """Called when the Button is clicked/pressed using a primary input (mouse1, return).

    Args:
        self: The instance that sent the event.
    """

    rules = f"""
    Button:
        content_style: dim
        alignment: [center, center]

        height: 1
        frame: [heavy, null, heavy, null]

        /idle:
            fill_style: '@ui.panel1'
            frame_style: ui.primary+1

        /hover:
            fill_style: '@ui.panel1+1'
            frame_style: ui.primary+1

        /selected:
            content_style: dim bold
            fill_style: '@ui.panel1+1'
            frame_style: ui.primary+1

        /active:
            fill_style: '@ui.primary+3'
            frame_style: 'ui.panel1-2'
            frame: frameless

        # Variants

        .big:
            height: 3

    {SEMANTIC_STYLES.format(base="success")}
    {SEMANTIC_STYLES.format(base="warning")}
    {SEMANTIC_STYLES.format(base="error")}
    """

    def __init__(
        self,
        content: str,
        on_submit: list[EventCallback] | None = None,
        **widget_args: Any,
    ) -> None:
        """Initializes the button.

        Args:
            content: The content the button should display.
            on_submit: A list of event callbacks to fire when the button submits.
            route: On click, it routes the application to a new page. Likely removed
                in the future.
        """

        super().__init__(**widget_args)

        self._has_timeout = False

        self.content = content
        self.width = sum(len(span) for span in self._parse_markup(self.content)) + 4

        self.on_submit = Event("button submit")
        for callback in on_submit or []:
            self.on_submit += callback

        self.bind("return", self._visual_submit)

    def _visual_submit(self) -> bool:
        """Animates the active state when the button is submitted using a keyboard."""

        if self._has_timeout:
            return False

        self.state_machine.apply_action("CLICKED")

        self._has_timeout = True
        self.app.timeout(
            150,
            lambda: (
                self.state_machine.apply_action("RELEASED_KEYBOARD"),
                setattr(self, "_has_timeout", False),  # type: ignore
            ),
        )

        return self.on_submit(self)

    def on_click(self, _: MouseAction, __: tuple[int, int]) -> bool:
        """Emits the submit event."""

        self.on_submit(self)

        return True

    def get_content(self) -> list[str]:
        return [self.content]



from typing import Any
from string import printable, punctuation, whitespace

from slate import Key

from ..enums import MouseAction
from .widget import Widget

PRINTABLE = [*printable]
DELIMITERS = {*whitespace, *punctuation}


class Field(Widget):
    """A text field."""

    style_map = Widget.style_map | {
        state: {"cursor": "", "placeholder": ""}
        for state in Widget.state_machine.states
    }

    rules = """
    Field:
        frame: [heavy, null, null, null]
        height: 1

        placeholder_style: 'dim italic'

        /idle|hover|active:
            frame_style: ui.primary-1

        /hover:
            fill_style: '@ui.panel1-2'

        /selected:
            content_style: ''
            frame_style: ui.primary-1
            cursor_style: '@ui.panel1+3'
    """

    def __init__(
        self,
        value: str = "",
        placeholder: str = "",
        multiline: bool = False,
        name: str | None = None,
        **widget_args,
    ) -> None:
        self.value = value
        self.placeholder = placeholder
        self.multiline = multiline
        self.name = name

        self.cursor = (0, 0)

        self._value_length = len(value)

        super().__init__(**widget_args)

    def move_cursor(self, x: int = 0, y: int = 0) -> None:
        """Moves the cursor by the given x and y coordinates."""

        self.cursor = min(self.cursor[0] + x, self._value_length), self.cursor[1] + y

    def on_click(self, _: MouseAction, __: tuple[int, int]) -> bool:
        """Allows the widget to be selected on click."""

        return True

    def handle_keyboard(self, key: Key) -> bool:
        binds = super().handle_keyboard(key)

        if key == "left":
            self.move_cursor(x=-1)
            return True

        if key == "right":
            self.move_cursor(x=1)
            return True

        x = self.cursor[0]
        left, right = self.value[:x], self.value[x:]

        if key == "backspace":
            if x == 0:
                return False

            self.value = left[:-1] + right
            self._value_length -= 1

            self.move_cursor(x=-1)
            return True

        if key == "ctrl-backspace":
            self.value = right
            change = len(left)

            self._value_length -= change
            self.move_cursor(x=-change)
            return True

        if key == "alt-backspace":
            distance = 0

            for distance, char in enumerate(reversed(left)):
                if char in DELIMITERS:
                    if distance == 0:
                        continue

                    break

            else:
                distance = len(left)

            self.value = left[:-distance] + right
            self.move_cursor(x=-distance)
            return True

        if key in PRINTABLE:
            self.value = left + str(key) + right
            self._value_length += 1

            self.move_cursor(x=1)
            return True

        return binds

    def serialize(self) -> dict[str, Any]:
        if self.name is None:
            raise ValueError(f"field {self!r} cannot be serialized without a name.")

        return {self.name: self.value}

    def get_content(self) -> list[str]:
        x = self.cursor[0]

        value = self.value or self.placeholder

        if not value:
            return [" " + self.styles["cursor"](" ") + "[/] "]

        style = self.styles["placeholder" if not self.value else "content"]

        left, right = value[:x], value[x + 1 :]

        cursor = " "

        if x < len(value):
            cursor = value[x]

        return [" " + style(left) + self.styles["cursor"](cursor) + style(right) + " "]



from ..enums import MouseAction
from .container import Tower


class Temp(Tower):
    def __init__(self, **widget_args) -> None:
        super().__init__(**widget_args)

        self._offset = None

    def on_click(self, _: MouseAction, position: tuple[int, int]) -> bool:
        self._offset = tuple(position[i] - self.position[i] for i in range(2))

        return True

    def on_release(self, *_) -> bool:
        self._offset = None

        return True

    def on_drag(self, _, position: tuple[int, int]) -> bool:
        if self._offset is None:
            return False

        self.position = tuple(position[i] - self._offset[i] for i in range(2))
        return True



from __future__ import annotations

from typing import Iterator, Iterable, Any
from slate import Span, Key
from .widget import Widget, _compute, handle_mouse_on_children
from ..enums import MouseAction, Direction, Alignment

__all__ = [
    "Container",
    "Tower",
    "Row",
]


def _align(alignment: Alignment, available: int) -> tuple[int, int]:
    """Returns offset & modulo result for alignment in the available space."""

    available = max(available, 0)

    if available == 1:
        return 0, 0

    if alignment == Alignment.CENTER:
        return divmod(available, 2)

    if alignment == Alignment.END:
        return available, 0

    return 0, 0


class Container(Widget):  # pylint: disable=too-many-public-methods
    """A widget that displays others based on some arrangement algorithm.

    See `Tower` and `Splitter` for the implementations you should use.
    """

    style_map = Widget.style_map | {
        "idle": {"scrollbar_y": "ui.secondary"},
        "hover": {"fill": Widget.style_map["idle"]["fill"]},
    }

    gap: int | float | None = None
    fallback_gap: int = 0

    _direction: Direction = Direction.VERTICAL

    def __init__(self, *children: Widget, **widget_args: Any) -> None:
        """Initializes a container.

        Args:
            *children: The children this widget should start with.
        """

        self.children: list[Widget] = []
        self.extend(children)

        super().__init__(**widget_args)

        self._should_layout = False
        self._layout_state: int = -1
        self._mouse_target: Widget | None = None
        self._hover_target: Widget | None = None
        self._outer_dimensions = (1, 1)

    @property
    def visible_children(self) -> list[Widget]:
        """Returns all children without the 'hidden' group set."""

        return [widget for widget in self.children if not "hidden" in widget.groups]

    @property
    def direction(self) -> Direction:
        """Returns and sets the current flow direction."""

        return self._direction

    @direction.setter
    def direction(self, new: Direction | str) -> None:
        """Sets the direction setting."""

        if isinstance(new, str):
            new = Direction(new)

        self._direction = new

    @property
    def selectable_count(self) -> int:
        return sum(child.selectable_count for child in self.children)

    def __iadd__(self, other: object) -> Container:
        """Adds a widget to the Container.

        Analogous to `.append(other)`.
        """

        if not isinstance(other, Widget):
            raise TypeError(
                f"Can only append widgets to containers, not {type(other)!r}."
            )

        self.append(other)
        return self

    def _is_fill(self, widget: Widget, horizontal: bool) -> bool:
        """Determines whether the widget is fill on the flow axis.

        Args:
            widget: The widget to test.
            horizontal: If set, flow direction is horizontal and we return
                `widget.is_fill_width()`, instead of `is_fill_height`.
        """

        if horizontal:
            return widget.is_fill_width()

        return widget.is_fill_height()

    def _compute_alignment_offsets(
        self,
        child: Widget,
        available_width: int,
        available_height: int,
        horizontal: bool,
    ) -> tuple[int, int, int]:
        """Computes x & y offsets to align by, based on alignment and flow direction.

        Args:
            child: The widget we are aligning. Used to get `computed_width` and
                `computed_height` as needed.
            available_width: The width the widget can be aligned within.
            available_height: The height the widget can be aligned within.
            horizontal: Whether flow direction is horizontal (True) or vertical (False).

        Returns:
            The x offset, y offset and the modulo of the operation that got either x
                offset or y offset, based on `horizontal`.
        """

        if horizontal:
            y, _ = _align(self.alignment[1], available_height - child.computed_height)
            x, extra = _align(self.alignment[0], available_width)

        else:
            x, _ = _align(self.alignment[0], available_width - child.computed_width)
            y, extra = _align(self.alignment[1], available_height)

        return x, y, extra

    def _compute_gap(self, available: int, count: int) -> tuple[int, int]:
        """Computes a gap based on available // count."""

        per_widget, extra = divmod(available, count or 1)

        return _compute(self.gap, per_widget), extra

    @property
    def selected(self) -> Widget | None:
        """Returns the currently selected widget."""

        if self._selected_index is None:
            return None

        return self.selectables[self._selected_index][0]

    @property
    def selectables(self) -> list[tuple[Widget, int]]:
        """Gets all selectable widgets and their inner indices.

        This is used in order to have a constant reference to all selectable indices
        within this widget.

        Returns:
            A list of tuples containing a widget and an integer each. For each widget
            that is withing this one, it is added to this list as many times as it has
            selectables. Each of the integers correspond to a selectable_index within
            the widget.

            For example, a Container with a Button, InputField and an inner Container
            containing 3 selectables might return something like this:

            ```
            [
                (Button(...), 0),
                (InputField(...), 0),
                (Container(...), 0),
                (Container(...), 1),
                (Container(...), 2),
            ]
            ```
        """

        _selectables: list[tuple[Widget, int]] = []

        for widget in self.children:
            if widget.selectable_count == 0:
                continue

            for i, (inner, _) in enumerate(widget.selectables):
                _selectables.append((inner, i))

        return _selectables

    def select(self, index: int | None = None) -> None:
        """Selects inner subwidget.

        Args:
            index: The index to select.

        Raises:
            IndexError: The index provided was beyond len(self.selectables).
        """

        if self.selectable_count == 0:
            return

        # Unselect all sub-elements
        for other in self.children:
            if other.selectable_count > 0:
                other.select(None)

        if index is not None:
            index = max(0, min(index, len(self.selectables) - 1))
            widget, inner_index = self.selectables[index]
            widget.select(inner_index)

        super().select(index)

    def serialize(self) -> dict[str, str]:
        data: dict[str, str] = {}

        for child in self.children:
            data.update(**child.serialize())

        return data

    def insert(self, index: int, widget: Widget) -> None:
        """Inserts a widget.

        Analogous to `list.insert`.
        """

        self.children.insert(index, widget)
        widget.parent = self
        self._should_layout = True

    def append(self, widget: Widget) -> None:
        """Adds a new widget setting its parent attribute to self.

        Analogous to `list.append`.
        """

        self.insert(len(self.children), widget)

    def extend(self, widgets: Iterable[Widget]) -> None:
        """Extends our children by the given iterable.

        Analogous to `list.extend`.
        """

        for widget in widgets:
            self.append(widget)

    def remove(self, widget: Widget) -> None:
        """Removes a widget from self, resetting its parent attribute to None.

        Analogous to `list.remove`.
        """

        self.children.remove(widget)
        widget.parent = None
        self._should_layout = True

    def pop(self, index: int) -> Widget:
        """Pops a widget from our children.

        Analogous to `list.pop`.

        Returns:
            The widget that was just removed.
        """

        widget = self.children[index]
        self.remove(widget)

        return widget

    def clear(self) -> None:
        """Removes all widgets.

        Analogous to `list.clear`.
        """

        for widget in self.children:
            self.remove(widget)

    def update_children(self, widgets: Iterable[Widget]) -> None:
        """Updates our children to the given iterable.

        Analogous to `list.update`, doesn't use `update` because it is already a widget
        method.
        """

        self.clear()
        self.extend(widgets)

    def replace(self, current: Widget, new: Widget, *, offset: int = 0) -> None:
        """Replaces `current` in children to `new`."""

        index = self.children.index(current)

        self.children[index + offset] = new
        new.parent = self

    def move_by(self, x: int, y: int) -> None:
        """Moves the widget (and all its children) to the given position."""

        super().move_by(x, y)

        for child in self.children:
            child.move_by(x, y)

    def arrange(  # pylint: disable=too-many-locals,too-many-statements,too-many-branches
        self, x: int, y: int
    ) -> None:
        """Arranges the widget's children according to its flow.

        Args:
            x: The origin's horizontal coordinate.
            y: The origin's vertical coordinate.
        """

        children = [child for child in self.visible_children]

        width = self._framed_width - self.has_scrollbar(1)
        height = self._framed_height - self.has_scrollbar(0)
        count = len(children) - len([child for child in children if child.is_static()])

        horizontal = self.direction == Direction.HORIZONTAL
        available = width if horizontal else height

        fills = 0

        for child in children:
            if self._is_fill(child, horizontal):
                fills += 1
                continue

            child.compute_dimensions(width, height)

            if child.is_static():
                continue

            if horizontal:
                available -= child.computed_width
            else:
                available -= child.computed_height

        if fills != 0:
            gap = self.fallback_gap
            gap_extra = 0

        else:
            # To avoid div by 0
            fills = 1

            gap, gap_extra = self._compute_gap(available, count)

            if gap * (count - 1) >= available:
                gap = self.fallback_gap
                gap_extra = 0

        available -= gap * (count - 1) + gap_extra

        fill_size, fill_extra = divmod(available, fills)

        this_gap = 0
        align_extra = 0

        for child in children:
            if self._is_fill(child, horizontal):
                this_fill = fill_size + (1 if fill_extra > 0 else 0)
                fill_extra -= 1

                if horizontal:
                    child.compute_dimensions(this_fill, height)

                else:
                    child.compute_dimensions(width, this_fill)

            if child.is_static():
                continue

            this_gap = gap + (1 if gap_extra > 0 else 0)
            gap_extra -= 1

            if horizontal:
                align_x, align_y, align_extra = self._compute_alignment_offsets(
                    child, this_gap, height, horizontal
                )

            else:
                align_x, align_y, align_extra = self._compute_alignment_offsets(
                    child, width, this_gap, horizontal
                )

            child.move_to(x + align_x, y + align_y)

            if horizontal:
                x += child.computed_width + this_gap + align_extra
                child.move_by(align_extra, 0)
            else:
                y += child.computed_height + this_gap + align_extra
                child.move_by(0, align_extra)

        if horizontal:
            self._outer_dimensions = (x, self.computed_height - this_gap - align_extra)
        else:
            self._outer_dimensions = (self.computed_width - this_gap - align_extra, y)

    def get_content(self) -> list[str]:
        """Calls our `arrange` method and returns a single empty line."""

        # layout_state = self._as_layout_state()

        # if layout_state != self._layout_state:
        start_x = self.position[0] + (self.frame.left != "")
        start_y = self.position[1] + (self.frame.top != "") + (self._clip_start or 0)

        self.arrange(start_x - self.scroll[0], start_y - self.scroll[1])
        # self._layout_state = layout_state

        return [""]

    def handle_keyboard(self, key: Key) -> bool:
        if self.selected is not None and self.selected.handle_keyboard(key):
            return True

        idx = self.selected_index or 0

        if key in ("left", "up", "shift-tab"):
            self.select(idx - 1)

        if key in ("right", "down", "tab"):
            self.select(idx + 1)

        return super().handle_keyboard(key)

    def handle_mouse(self, action: MouseAction, position: tuple[int, int]) -> bool:
        result, selection, mouse_target, hover_target = handle_mouse_on_children(
            action,
            position,
            self._mouse_target,
            self._hover_target,
            self.visible_children,
        )

        self._mouse_target = mouse_target
        self._hover_target = hover_target

        if result:
            if selection is not None:
                self.select(selection)
            return True

        # There was a click, but it didn't get handled; likely clicked on a
        # non-selectable, or empty space.
        if "click" in action.value:
            self.select(None)

        return super().handle_mouse(action, position)

    def drawables(self) -> Iterator[Widget]:
        # TODO: Implement child-clipping

        yield self

        for widget in self.children:
            yield from widget.drawables()

    def build(
        self, *, virt_width: int | None = None, virt_height: int | None = None
    ) -> list[tuple[Span, ...]]:
        my_virt_width, my_virt_height = self._outer_dimensions

        return super().build(virt_width=my_virt_width, virt_height=my_virt_height)


class Tower(Container):
    """Arranges widgets in a vertical tower.

    ```
    +----------------+ <- alignment=("center", "center")
    |                |
    |     +---+      |
    |     |   | <------- width=0.5
    |     +---+      |
    |   +--------+   |
    |   |        | <---- width=1.0
    |   +--------+   |
    |    +-----+     |
    |    |     | <------ width=7
    |    +-----+     |
    |                |
    +----------------+
    ```
    """

    direction = Direction.VERTICAL


class Row(Container):
    """Arranges widgets in a horizontal row.

    ```
    +----------------------------+ <- alignment=("start", "center")
    |                            |
    |      +--------++----------+|
    |+----+|        ||          ||
    ||    ||        ||          ||
    |+----+|        ||          ||
    |  ^   +--------++----------+|
    |  |          ^         ^    |
    +--|----------|---------|----+
       |_ width=6 |         |
                  |_ width=0.3
                            |_ width=0.3
    ```
    """

    direction = Direction.HORIZONTAL



# pylint: disable=too-many-lines

from __future__ import annotations

import re
from dataclasses import dataclass
from threading import Thread
from time import perf_counter, sleep
from typing import Any, Callable, Iterator, Iterable, Type, overload
from types import TracebackType
from yaml import safe_load

from slate import Terminal, getch, Event, terminal as slt_terminal, feed, Key

from .widgets import Widget, handle_mouse_on_children, Container
from .enums import MouseAction
from .state_machine import deep_merge
from .style_map import StyleMap

__all__ = [
    "load_rules",
    "Selector",
    "Application",
    "Page",
]

DEFAULT_RULES = """
.fill:
    width: null
    height: null

.w-fill:
    width: null

.h-fill:
    height: null

.start:
    alignment: [start, start]

.center:
    alignment: [center, center]

.end:
    alignment: [end, end]

.of-scroll:
    overflow: [scroll, scroll]

.of-auto:
    overflow: [auto, auto]

.of-hide:
    overflow: [hide, hide]
"""


def _parse_mouse_input(key: Key) -> tuple[MouseAction, tuple[int, int]] | None:
    inp = str(key)

    if not inp.startswith("mouse:"):
        return None

    # TODO: This ignores stacked events and only handles the last one. Shouldn't be an
    # issue, but look out.
    inp = inp.rsplit("mouse:", maxsplit=1)[-1]

    action, position = inp.split("@")
    parts = position.split(";")

    if len(parts) != 2:
        return None

    return MouseAction(action), (int(parts[0]), int(parts[1]))


def _flatten(widget: Widget) -> list[Widget]:
    items = [widget]

    for child in widget.drawables():
        if child is widget:
            continue

        items.extend(_flatten(child))

    return items


def load_rules(source: str) -> dict[str, dict[str, Any]]:
    """Loads a set of rule declarations from YAML.

    Nested get inserted into the parent's key, so:

    ```yaml
    Button:
        height: 1
        frame: lightvertical

        /idle:
            content_style: primary-1
    ```

    would become:

    ```json
    {
        "Button": {
            "height": 1,
            "frame": "lightvertical",
        },
        "Button/idle": {
            "content_style": "primary-1",
        }
    }
    ```

    You can customize where the insertion takes place by marking it with an `&`. For
    example:

    ```yaml
    .fill:
        height: 1

        Progressbar&:
            height: 3
    ```

    becomes:

    ```json
    {
        ".fill": {
            "height": 1,
        },
        "Progressbar.fill": {
            "height": 3,
        }
    }
    ```

    Returns:
        A flattened dictionary of `{selector: {rules...}}`.
    """

    source_data = safe_load(source)

    if source_data is None:
        return {}

    outer = {}

    def _flatten(data: dict[str, Any], prefix: str = "") -> dict[str, Any]:
        inner = {}

        for key, value in data.items():
            if isinstance(value, dict):
                if key.startswith((">", "*>")):
                    key = " " + key

                if "&" not in key:
                    key = f"&{key}"

                key = key.replace("&", prefix)

                for part in key.split(","):
                    outer[part.lstrip()] = _flatten(value, prefix=part)

                continue

            if isinstance(value, list):
                value = tuple(value)

            inner[key] = value

        return inner

    _flatten(source_data)

    return outer


RE_QUERY = re.compile(
    r"^([A-Z][\w\|]+)?(#[a-z0-9@\-]+)?((?:\.[a-z0-9@\-]+)*)?(\/[\w\|]+)?$"
)


@dataclass(frozen=True)
class Selector:  # pylint: disable=too-many-instance-attributes
    """The object used to query widgets.

    Basic syntax is:

        WidgetType#id.group1.group2/state1|state2

    ...where `WidgetType` is necessary if an id is given, everything else is optional.

    You can also test for hierarchy, e.g.:

        # Direct hierarchy (ParentType > WidgetType.class1)
        ParentType > WidgetType.class1

        # Indirect hierarchy (ParentType > OtherParent > WidgetType.class1)
        ParentType *> WidgetType.class1

    Finally, you can use `*` to indicate 'any' in hierarchal matches:

        # WidgetType#id that is the child of any widget (all of them, so don't do this)
        * > WidgetType#id

        # Any widget whos direct parent is WidgetType#id
        WidgetType#id > *
    """

    query: str

    elements: tuple[str, ...] = tuple()
    eid: str | None = None
    groups: tuple[str, ...] = tuple()
    states: tuple[str, ...] | None = None

    direct_parent: Selector | None = None
    indirect_parent: Selector | None = None
    forced_score: int | None = None

    def __str__(self) -> str:
        if self.direct_parent is not None:
            return f"{self.direct_parent} > {self.query}"

        if self.indirect_parent is not None:
            return f"{self.indirect_parent} *> {self.query}"

        return self.query

    @classmethod
    def parse(
        cls,
        query: str,
        direct_parent: Selector | None = None,
        indirect_parent: Selector | None = None,
        forced_score: int | None = None,
    ) -> Selector:
        """Parses a query into a selector.

        Args:
            query: The query to parse.
            direct_parent: Direct parent to test for, obtained by parsing
                `direct_parent > query`.
            indirect_parent: Direct parent to test for, obtained by parsing
                `indirect_parent *> query`.
            forced_score:A score to use when a selector matches the rule, instead of the
                score the selector calculates.

        Returns:
            The obtained selector.
        """
        # TODO: Add support for multi-hierarchy
        if " > " in query:
            left, right = query.split(" > ")

            return Selector.parse(
                right, direct_parent=Selector.parse(left), forced_score=forced_score
            )

        if " *> " in query:
            left, right = query.split(" *> ")

            return Selector.parse(
                right, indirect_parent=Selector.parse(left), forced_score=forced_score
            )

        if query == "*":
            return cls(
                query,
                ("Widget",),
                forced_score=forced_score,
                direct_parent=direct_parent,
                indirect_parent=indirect_parent,
            )

        mtch = RE_QUERY.match(query)

        if mtch is None:
            raise ValueError(
                f"incorrect syntax {query!r}, use 'WidgetType#id.class/type'"
            )

        elements_str, eid, groups_str, states = mtch.groups()
        elements_str = elements_str or ""

        elements = tuple(elements_str.split("|"))
        eid = (eid or "").lstrip("#") or None
        groups = tuple(groups_str.split(".")[1:])

        if states is not None:
            states = tuple(states.lstrip("/").split("|"))

        return cls(
            query=query,
            elements=elements,
            eid=eid,
            groups=groups,
            states=states,
            direct_parent=direct_parent,
            indirect_parent=indirect_parent,
            forced_score=forced_score,
        )

    def _match_parent(self, widget: Widget | "Page", direct: bool = True) -> int:
        """Walks up the widget's parent tree and tries to match each.

        Args:
            widget: The widget whos parents we are looking at.
            direct: If set, we will only walk up for one node, and use `direct_parent`
                instead of `indirect_parent`.
        """

        if widget.parent is None:
            return 0

        parent = widget.parent

        while hasattr(parent, "parent") and parent.parent is not None:
            if direct:
                # Both are checked to be non-null at call site.
                return self.direct_parent.matches(parent)  # type: ignore

            if score := self.indirect_parent.matches(parent):  # type: ignore
                return score

            parent = widget.parent

        return 0

    def matches(  # pylint: disable=too-many-return-statements, too-many-branches
        self, widget: Widget | "Page"
    ) -> int:
        """Determines how well this selector matches the widget.

        Returns:
            A score calculated as:

                (
                    (eid_matches * 1000)
                    + (group_matches * 500 + len(groups) * 100)
                    + (state_matches * 250)
                    + (type_matches * 100)
                )

            When optional parts are ommitted (like groups), they are excluded from the
            calculation. However, when these are provided the widget _must_ have
            matching attributes, otherwise the function will return 0.
        """

        if self.query == "*":
            return 10

        if type(widget).__name__ not in self.elements and self.elements != ("",):
            return 0

        if self.forced_score is not None:
            return self.forced_score

        score = 100

        if self.direct_parent is not None:
            if (scr := self._match_parent(widget)) == 0:
                return 0

            score += scr

        elif self.indirect_parent is not None:
            if (scr := self._match_parent(widget, direct=False)) == 0:
                return 0

            score += scr

        if self.eid is not None:
            if isinstance(widget, Widget) and widget.eid == self.eid:
                score += 1000

            else:
                return 0

        if self.groups:
            if isinstance(widget, Widget) and all(
                group in widget.groups for group in self.groups
            ):
                score += 500 + 100 * len(self.groups)

            else:
                return 0

        if self.states is not None:
            if isinstance(widget, Widget) and any(
                widget.state.endswith(state) for state in self.states
            ):
                score += 250

            else:
                return 0

        return score


BuilderType = Callable[..., "Page"]


class Page:  # pylint: disable=too-many-instance-attributes
    """A Page of an application.

    It contains some children, and a set of rules it applies to them.
    """

    parent: "Application"
    children: list[Widget]
    _rules: dict[Selector, tuple[dict[str, Any], dict[str, Any]]]

    def __init__(
        self,
        *children: Widget,
        title: str = "",
        route_name: str = "/",
        builder: BuilderType | None = None,
        rules: str = "",
    ) -> None:
        if not route_name.startswith("/"):
            raise ValueError(f"route names must start with a slash, got {route_name!r}")

        self.title = title
        self.route_name = route_name
        self._children: list[Widget] = []
        self._rules = {}
        self._encountered_types: list[type] = []
        self._builder = builder
        self._rules_changed = True

        self._init_rules = rules

        self.extend(children)

    def __iadd__(self, widget: Any) -> Page:
        """Shorthand for `.append(widget)`."""

        if not isinstance(widget, Widget):
            raise TypeError(f"Can only add Widgets (not {widget!r}) to Page")

        self.append(widget)

        return self

    def __iter__(self) -> Iterator[Widget]:
        """Shorthand for `iter(._children)`."""

        return iter(self._children)

    def __len__(self) -> int:
        """Shorthand for `len(._children)`."""

        return len(self._children)

    @overload
    def __getitem__(self, item: int) -> Widget:
        ...

    @overload
    def __getitem__(self, item: slice) -> list[Widget]:
        ...

    def __getitem__(self, item):
        """Shorthand for `._children[item]`."""

        return self._children[item]

    def _init_widget(self, widget: Widget) -> None:
        """Initializes a widget.

        This loads the widget's rules if it hasn't encountered the same type before,
        and sets its parent to self.
        """

        if type(widget) not in self._encountered_types:
            for child in widget.drawables():
                self.load_rules(child.rules, score=None)
                self._encountered_types.append(type(child))

        widget.parent = self

    def build(self, *args: Any, **kwargs: Any) -> Page:
        """Applies the builder function.

        Might be removed in the future.
        """

        if self._builder is None:
            if len(args) + len(kwargs):
                raise ValueError(
                    f"Page {self!r} takes no arguments, got {args=!r}, {kwargs!r}."
                )

            return self

        return self._builder(*args, **kwargs)

    def append(self, widget: Widget) -> None:
        """Initializes & adds a widget to the Page.

        Analogous to `list.append`.
        """

        self._init_widget(widget)
        self._children.append(widget)

    def extend(self, widgets: Iterable[Widget]) -> None:
        """Extends the page by the given widgets.

        Analogous to `list.extend`.
        """

        for widget in widgets:
            self._init_widget(widget)
            self.append(widget)

    def insert(self, index: int, widget: Widget) -> None:
        """Inserts a widget into the page.

        Analogous to `list.insert`.
        """

        self._init_widget(widget)
        self._children.insert(index, widget)

    def remove(self, widget: Widget) -> None:
        """Removes a widget from the page.

        Analogous to `list.remove`.
        """

        self._init_widget(widget)
        self._children.remove(widget)

    def pop(self, index: int) -> Widget:
        """Pops a widget from the page.

        Analogous to `list.pop`.
        """

        widget = self._children.pop(index)

        return widget

    def clear(self) -> None:
        """Removes all widgets.

        Analogous to `list.clear`.
        """

        for widget in self._children:
            self.remove(widget)

    def update(self, widgets: Iterable[Widget]) -> None:
        """Clears all widgets then adds everything given.

        Analogous to `list.update`.
        """

        self.clear()
        self.extend(widgets)

    def load_rules(
        self,
        rules: str | None = None,
        score: int | None = None,
        load_init_rules: bool = False,
    ) -> None:
        """Loads the given YAML rules.

        Args:
            rules: The YAML to load.
            score: A score to use when a selector matches the rule, instead of the score
                the selector calculates.
            load_init_rules: If set, will ignore rules and use the rules passed in
                during `__init__`.

        Either `rules` or `load_init_rules` must be set.
        """

        if rules is None and not load_init_rules:
            raise TypeError("must provide rules to load.")

        if load_init_rules:
            rules = self._init_rules

        assert rules is not None

        for selector, rule in load_rules(rules).items():
            self.rule(selector, **rule, score=score)

    def apply_rules(self) -> bool:
        """Applies the page's rules to the widgets.

        Returns:
            Whether any rules were applied, e.g. whether there was any change.
        """

        drawables: list[Widget] = []

        for widget in self._children:
            drawables.extend(widget.drawables())

        applicable_rules = None

        rules_changed = self._rules_changed

        for widget in drawables:
            new_attrs: dict[str, Any] = {}
            new_style_map = StyleMap()

            if not rules_changed and not widget.query_changed():
                continue

            applicable_rules = [
                (sel, score, rule)
                for sel, rule in self._rules.items()
                if (score := sel.matches(widget)) != 0
            ]

            for sel, score, (attrs, style_map) in sorted(
                applicable_rules, key=lambda item: item[1]
            ):
                new_attrs.update(**attrs)
                new_style_map |= style_map

            widget.update(new_attrs, new_style_map)

        self._rules_changed = False

        return applicable_rules is not None

    def find_all(
        self, query: str | Selector, scope: Container | None = None
    ) -> Iterator[Widget]:
        """Finds all widgets in the page matching the given query."""

        selector: Selector

        if isinstance(query, Selector):
            selector = query
        else:
            selector = Selector.parse(query)

        children = self._children if scope is None else scope.children

        for widget in children:
            for child in widget.drawables():
                if selector.matches(child):
                    yield child

    def find(
        self, query: str | Selector, scope: Container | None = None
    ) -> Widget | None:
        """Finds the first widget in the page matching the given query."""

        for widget in self.find_all(query, scope):
            return widget

        return None

    def rule(
        self, query: str | Selector, score: int | None = None, **rules: str
    ) -> Selector:
        """Creates a new rule that matches query, and applies rules on matches.

        Args:
            query: The query that will be matched against.
            score: A score to use when a selector matches the rule, instead of the score
                the selector calculates.
            rules: The rules to apply to every matching widget.

        Returns:
            The selector used for matching.
        """

        style_map = {}
        attrs = {}

        self._rules_changed = True

        for key, value in rules.copy().items():
            if not key.endswith("_style"):
                attrs[key] = value
            else:
                style_map[key[: -len("_style")]] = value

        if isinstance(query, Selector):
            selector = query
        else:
            selector = Selector.parse(query, forced_score=score)

        if selector not in self._rules:
            self._rules[selector] = (attrs, style_map)
            return selector

        # Merge existing rules for the same selector
        old_attrs, old_style_map = self._rules[selector]
        deep_merge(old_attrs, attrs)
        deep_merge(old_style_map, style_map)

        return selector


class Application(Page):  # pylint: disable=too-many-instance-attributes
    """A page to rule them all.

    The primary interface to run Celadon applications. It maintains a list of pages,
    between which it can navigate using routing. Only one page can be displayed at a
    time, but it is possible to create constant overlays by adding widgets directly to
    the Application instance, instead of a page.
    """

    children: list[Widget]  # For constantly visible, overlay-type widgets

    _pages: list[Page]
    _page: Page | None

    def __init__(
        self, title: str, framerate: int = 60, terminal: Terminal | None = None
    ) -> None:
        """Initializes the Application.

        Args:
            name: The name of the app. Used for the terminal's title bar.
            framerate: The target framerate. Note that this isn't very closely
                maintained.
            terminal: A terminal instance to use.
        """

        super().__init__(title=title, route_name="/")

        self.on_frame_drawn = Event("frame drawn")
        self.on_page_added = Event("page added")
        self.on_page_changed = Event("page changed")

        self._pages = []
        self._page = None
        self._mouse_target: Widget | None = None
        self._hover_target: Widget | None = None
        self._framerate = framerate
        self._terminal = terminal or slt_terminal

        self._is_running = False
        self._is_paused = False
        self._raised: Exception | None = None
        self._timeouts: list[tuple[Callable[[], Any], int | float]] = []

        _ = self.terminal.foreground_color
        _ = self.terminal.background_color

    def __enter__(self) -> Application:
        Widget.app = self

        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        if exc is not None:
            raise exc

        self.run()

    def __iadd__(self, page: Any) -> Application:
        """Shorthand for `.append(page)`."""

        if not isinstance(page, Page):
            raise TypeError(f"Can only add pages (not {page!r}) to App")

        self.append(page)

        return self

    def __iter__(self) -> Iterator[Page]:  # type: ignore
        """Shorthand for `iter(._pages)`."""

        return iter(self._pages)

    def _draw_loop(self) -> None:  # pylint: disable=too-many-locals
        """The display & timing loop of the Application, run as a thread."""

        frametime = 1 / self._framerate

        clear = self._terminal.clear
        write = self._terminal.write
        draw = self._terminal.draw

        on_frame_drawn = self.on_frame_drawn

        try:
            while self._is_running:
                if self._is_paused:
                    sleep(frametime)

                start = perf_counter()
                clear()

                self.apply_rules()

                for widget in [*self._page, *self._children]:  # type: ignore
                    widget.compute_dimensions(
                        self._terminal.width, self._terminal.height
                    )

                    for child in widget.drawables():
                        origin = child.position

                        for i, line in enumerate(child.build()):
                            write(line, cursor=(origin[0], origin[1] + i))

                draw()

                if on_frame_drawn:
                    on_frame_drawn()
                    on_frame_drawn.clear()

                elapsed = perf_counter() - start

                if elapsed < frametime:
                    sleep(frametime)

                eliminated = []

                for i, (callback, timeout) in enumerate(self._timeouts):
                    timeout -= elapsed * 1000

                    if timeout <= 0:
                        callback()

                        eliminated.append(i)
                        continue

                    self._timeouts[i] = (callback, timeout)

                for i in eliminated:
                    self._timeouts.pop(i)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.stop()
            self._raised = exc

    @property
    def page(self) -> Page | None:
        """Returns the current page."""

        return self._page

    @property
    def terminal(self) -> Terminal:
        """Returns the in-use terminal instance."""

        return self._terminal

    def timeout(self, delay_ms: int, callback: Callable[[], Any]) -> None:
        """Sets up a non-blocking timeout.

        Args:
            delay_ms: The delay (in milliseconds) before the callback is executed.
            callback: The callback to execute when the delay is up.
        """

        self._timeouts.append((callback, delay_ms))

    def find_all(
        self, query: str | Selector, scope: Container | None = None
    ) -> Iterator[Widget]:
        selector = query

        if isinstance(query, str):
            selector = Selector.parse(query)

        assert isinstance(selector, Selector)

        if self.page is not None:
            yield from self.page.find_all(selector, scope)

        yield from super().find_all(query, scope)

    def rule(
        self, query: str | Selector, score: int | None = None, **rules: str
    ) -> Selector:
        if isinstance(query, Selector):
            selector = query

        else:
            selector = Selector.parse(query, forced_score=score)

        for page in self._pages:
            page.rule(selector, score=score, **rules)

        super().rule(selector, score=score, **rules)

        return selector

    def append(self, page: Page) -> None:  # type: ignore # pylint: disable=arguments-renamed
        """Adds a page."""

        if page.title is None or page.route_name is None:
            raise ValueError(
                "Pages must have both `title` and `route_name`,"
                f" got {page.title=!r}, {page.route_name=!r}"
            )

        rules: dict[str, Any] = {}

        # God compelled me to write this,,,
        # May he compel me to clean  it up.
        #
        for key, value in self._rules.items():
            rules.update(
                **{
                    str(key): {
                        **value[0],
                        **{sel + "_style": val for sel, val in value[1].items()},
                    }
                }
            )

        for selector, rule in rules.items():
            page.rule(selector, **rule)

        page.parent = self

        self._pages.append(page)
        page.load_rules(load_init_rules=True)
        self.on_page_added()

        if self._mouse_target is None and len(page) > 0:
            self._mouse_target = page[0]

    def pin(self, widget: Widget) -> None:
        """Pins a widget to the application.

        Pinning means the app will always be displayed & on top, regardless of the
        current page.

        Analogous to `Page.append(self, widget)`.
        """

        super().append(widget)

    def apply_rules(self) -> bool:
        page_applied = False

        if self._page is not None:
            page_applied |= self._page.apply_rules()

        return page_applied | super().apply_rules()

    def route(self, destination: str) -> None:
        """Routes to a new page."""

        for page in self._pages:
            if page.route_name == destination:
                break

        else:
            raise ValueError(f"No page with route {destination!r}.")

        self._page = page
        self.on_page_changed()

        if page.route_name == "/":
            self._terminal.set_title(f"{self.title}")

        else:
            self._terminal.set_title(f"{self.title} - {page.title}")

        self.apply_rules()

    def process_input(self, inp: Key) -> bool:
        """Processes input.

        Returns:
            Whether the input could be handled.
        """

        event = _parse_mouse_input(inp)

        if event is None:
            if self._mouse_target is not None:
                return self._mouse_target.handle_keyboard(inp)

            return False

        result, *_, hover_target = handle_mouse_on_children(
            *event,
            self._mouse_target,
            self._hover_target,
            reversed([*self._page, *self._children]),  # type: ignore
        )

        # We need to keep (not update) `_mouse_target` to handle keyboard inputs
        self._hover_target = hover_target

        if result:
            return True

        return False

    def run(self) -> None:
        """Runs the application.

        This function will block until the application halts.
        """

        self._is_running = True

        if self._page is None:
            self.route("/")

        self.load_rules(DEFAULT_RULES)
        terminal = self._terminal

        with terminal.report_mouse(), terminal.no_echo(), terminal.alt_buffer():
            thread = Thread(name=f"{self.title}.draw", target=self._draw_loop)
            thread.start()

            while self._is_running:
                inp = getch()

                if self._is_paused:
                    continue

                if inp == "ctrl-c":
                    self.stop()
                    break

                try:
                    self.process_input(inp)

                except Exception as exc:  # pylint: disable=broad-exception-caught
                    self.stop()
                    self._raised = exc
                    break

            thread.join()

        self._terminal.set_title(None)

        if self._raised is not None:
            raise self._raised

    # TODO: Expand this to clear screen (wait for frame to finish)
    def pause(self) -> None:
        """Pauses the app."""

        self._is_paused = True

    def resume(self) -> None:
        """Resumes the app."""

        self._is_paused = False

    def stop(self) -> None:
        """Stops the application."""

        feed(chr(3))
        self._is_running = False



from __future__ import annotations

from copy import deepcopy
from typing import Any

from slate import Event

__all__ = [
    "StateMachine",
    "deep_merge",
]


def deep_merge(
    dict1: dict[Any, Any], dict2: dict[Any, Any], path: list[str] | None = None
):
    """Deep merges dict1 into dict2.

    Source: https://stackoverflow.com/a/7205107
    """

    if path is None:
        path = []

    for key in dict2:
        val2 = dict2[key]

        if key in dict1:
            val1 = dict1[key]

            if isinstance(val1, dict) and isinstance(val2, dict):
                deep_merge(val1, val2, path + [str(key)])

            else:
                dict1[key] = val2

        else:
            dict1[key] = val2

    return dict1


class StateMachine:
    """A state manager to manage widget state."""

    on_change: Event
    """Called when the state changes.

    Args:
        state: The new (changed-to) state.
    """

    def __init__(
        self, states: tuple[str, ...], *, transitions: dict[str, dict[str, str]]
    ) -> None:
        self.on_change = Event("State Changed")

        self.states = states
        self._transitions = transitions

        self._state = states[0]
        self._substate = "/"

    def __call__(self) -> str:
        """Returns the current state, including substate."""

        return self._state + (self._substate if self._substate != "/" else "")

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} state: {self()!r}>"

    def __contains__(self, item: object) -> bool:
        return item in self.states

    def __getitem__(self, index: int) -> str:
        return self.states[index]

    def copy(
        self,
        states: tuple[str, ...] | None = None,
        add_states: tuple[str, ...] | None = None,
        transitions: dict[str, dict[str, str]] | None = None,
        add_transitions: dict[str, dict[str, str]] | None = None,
    ) -> StateMachine:
        """Creates a copy of the state machine and modifies it according to the args.

        Args:
            states: Overwrites the original machine's states.
            add_states: Adds to the original machine's states.
            transitions: Overwrites the original machine's transitions.
            add_transitions: Adds to the original machine's transitions.
        """

        if states is not None:
            new_states = states

        elif add_states is not None:
            new_states = tuple((*self.states, *add_states))

        else:
            new_states = self.states

        if transitions is not None:
            new_transitions = transitions

        elif add_transitions is not None:
            new_transitions = deep_merge(deepcopy(self._transitions), add_transitions)

        else:
            new_transitions = deepcopy(self._transitions)

        return StateMachine(new_states, transitions=new_transitions)

    def apply_action(self, action: str) -> bool:
        """Applies some action to the state manager.

        Actions are strings that are used for updating state.

        Returns:
            Whether the given action resulted in a state change.

        Args:
            action: An action name.
        """

        if action.startswith("SUBSTATE_"):
            transitions = self._transitions[self._substate]

            substate = transitions.get(action)

            if substate is None:
                return False

            self._substate = substate
            return True

        transitions = self._transitions.get(self._state, {})

        state = transitions.get(action)

        # No defined transition from the current state by the action
        if state is None:
            return False

        self._state = state

        self.on_change(state)
        return True



from __future__ import annotations

from copy import deepcopy
from .state_machine import deep_merge


class StyleMap(dict):
    """A dictionary that is deep-merged when using the | (or) operator."""

    def __or__(self, other: object) -> StyleMap:
        if not isinstance(other, (dict, StyleMap)):
            raise TypeError(
                "Can only merge a StyleMap with a dict or another StyleMap,"
                f" not {type(other)!r}."
            )

        # if "*" in other:
        #     for key in

        return StyleMap(deep_merge(deepcopy(self), other))



# pylint: disable=too-few-public-methods

from __future__ import annotations

from typing import Type, Tuple

__all__ = [
    "Frame",
    "CharTuple",
    "get_frame",
]

CharTuple = Tuple[str, str, str, str]


class Frame:  # pylint: disable=too-many-instance-attributes
    """A set of characters to wrap a widget by."""

    descriptor: tuple[str, str, str] | None = None
    """A list of strings that describes the frame's borders & corners."""

    borders: CharTuple = ("", "", "", "")
    """Left, top, right and bottom border characters."""

    corners: CharTuple = ("", "", "", "")
    """Left-top, right-top, right-bottom and left-bottom border characters."""

    scrollbars: tuple[tuple[str, str], tuple[str, str]]
    """The characters to use for scrollbars (horizontal, vertical).

    It is stored in the order of (start_corner, thumb, rail, end_corner).
    """

    width: int
    """Width of the left and right borders combined."""

    height: int
    """Height of the top and bottom borders combined."""

    outer_horizontal: bool = False
    """Use the parent's fill color (instead of our own) when drawing the sides."""

    outer_vertical: bool = False
    """Use the parent's fill color (instead of our own) when drawing the top & bottom."""

    outer_corner: bool = False
    """Use the parent's fill color (instead of our own) when drawing the corners."""

    def __init__(self) -> None:
        if self.descriptor is not None:
            self.borders, self.corners = self._parse_descriptor()

        assert self.borders is not None and self.corners is not None

        self.width = len(self.borders[0] + self.borders[2])
        self.height = (self.borders[1] != "") + (self.borders[3] != "")

        self.left, self.top, self.right, self.bottom = self.borders
        (
            self.left_top,
            self.right_top,
            self.right_bottom,
            self.left_bottom,
        ) = self.borders

    def _parse_descriptor(self) -> tuple[CharTuple, CharTuple]:
        """Parses the descriptor into tuples of chartuples."""

        assert self.descriptor is not None
        top, middle, bottom = self.descriptor  # pylint: disable=unpacking-non-sequence

        left_top, top, *_, right_top = list(top)
        left, *_, right = list(middle)
        left_bottom, bottom, *_, right_bottom = list(bottom)

        return (
            (left, top, right, bottom),
            (left_top, right_top, right_bottom, left_bottom),
        )

    @property
    def name(self) -> str:
        """Return the frame class' name."""

        return type(self).__name__

    @classmethod
    def compose(
        cls, sides: tuple[Type[Frame], Type[Frame], Type[Frame], Type[Frame]]
    ) -> Frame:
        """Creates a frame object with borders coming from other frames.

        Args:
            sides: Sides to source from, read in the order left, top, right, bottom.

        Returns:
            A plain Frame object (so no custom name or documentation preview) with the
            given sides.
        """

        borders = []

        frame = cls()

        for i, side in enumerate(sides):
            borders.append(side().borders[i])

        frame.borders = tuple(borders)  # type: ignore
        frame.__init__()  # type: ignore # pylint: disable=unnecessary-dunder-call

        return frame


def get_frame(name: str) -> Type[Frame]:
    """Gets a frame by its name.

    Args:
        name: A case-insensitive frame name.

    Returns:
        The frame class matching the given name.

    Raises:
        ValueError: No frame found with the given name.
    """

    if name is None:
        return Frameless

    frame = {key.lower(): value for key, value in globals().items()}.get(name.lower())

    if frame is not None and issubclass(frame, Frame):
        return frame

    raise ValueError(f"No frame defined with name {name!r}.")


def add_frame_preview(cls: Type[Frame]):
    """Adds a documentation preview to the given class."""

    if cls.descriptor is None:
        return cls

    if cls.__doc__ is None:
        cls.__doc__ = ""

    cls.__doc__ += f"""

Preview:

```
{cls.descriptor[0]}
{cls.descriptor[1]}
{cls.descriptor[2]}
```
"""
    return cls


@add_frame_preview
class ASCII_X(Frame):  # pylint: disable=invalid-name
    """A frame of ASCII characters, with X-s in the corners."""

    descriptor = (
        "X---X",
        "|   |",
        "X---X",
    )

    scrollbars = (("-", "#"), ("|", "#"))


@add_frame_preview
class Light(Frame):
    """A frame with a light outline."""

    descriptor = (
        "┌───┐",
        "│   │",
        "└───┘",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class LightVertical(Frame):
    """A frame with a light outline."""

    borders = ("│", "", "│", "")
    corners = ("", "", "", "")

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class Heavy(Frame):
    """A frame with a heavy outline."""

    descriptor = (
        "┏━━━┓",
        "┃   ┃",
        "┗━━━┛",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class Rounded(Frame):
    """A frame with a light outline."""

    descriptor = (
        "╭───╮",
        "│   │",
        "╰───╯",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class Double(Frame):
    """A frame with a double-lined outline."""

    descriptor = (
        "╔═══╗",
        "║   ║",
        "╚═══╝",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class Dashed(Frame):
    """A frame with a dashed outline."""

    descriptor = (
        "┌╌╌╌┐",
        "╎   ╎",
        "└╌╌╌┘",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class HeavyDashed(Frame):
    """A frame with a dashed outline."""

    descriptor = (
        "┏╍╍╍┓",
        "╏   ╏",
        "┗╍╍╍┛",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class Padded(Frame):
    """A frame of spaces."""

    descriptor = (
        "   ",
        "   ",
        "   ",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class Frameless(Frame):
    """A frame of nothing."""

    borders = ("", "", "", "")
    corners = ("", "", "", "")

    scrollbars = ((" ", "▅"), (" ", "█"))


@add_frame_preview
class HorizontalOuter(Frame):
    """A horizontal McGugan box.

    https://www.willmcgugan.com/blog/tech/post/ceo-just-wants-to-draw-boxes/
    """

    descriptor = (
        "▁▁▁▁▁",
        "▎ x 🮇",
        "▔▔▔▔▔",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))

    outer_horizontal = True
    outer_corner = True


@add_frame_preview
class VerticalOuter(Frame):
    """A vertical McGugan box.

    https://www.willmcgugan.com/blog/tech/post/ceo-just-wants-to-draw-boxes/
    """

    descriptor = (
        "🮇▔▔▔▔▔▎",
        "🮇  x  ▎",
        "🮇▁▁▁▁▁▎",
    )

    scrollbars = ((" ", "▅"), (" ", "█"))

    outer_vertical = True
    outer_corner = True



from celadon import StateMachine


def get_state_machine():
    return StateMachine(
        states=(
            "idle",
            "hover",
            "selected",
            "active",
        ),
        transitions={
            "/": {
                "SUBSTATE_ENTER_BLUR": "/blur",
            },
            "/blur": {
                "SUBSTATE_EXIT_BLUR": "/",
            },
            "idle": {
                "HOVERED": "hover",
                "SELECTED": "selected",
                "CLICKED": "active",
                "SINK": "sinkhole",
            },
            "hover": {
                "UNHOVERED": "idle",
                "SELECTED": "selected",
                "CLICKED": "active",
            },
            "selected": {
                "UNSELECTED": "idle",
                "CLICKED": "active",
            },
            "active": {
                "RELEASED": "idle",
            },
            "sinkhole": {},
        },
    )


def test_state_machine_action_changes_state():
    state = get_state_machine()

    assert state() == "idle"

    state.apply_action("HOVERED")
    assert state() == "hover"

    state.apply_action("UNHOVERED")
    assert state() == "idle"

    assert not state.apply_action("NOT_AN_ACTION")

    state.apply_action("SINK")
    assert state() == "sinkhole"
    assert not state.apply_action("CLICKED")


def test_state_machine_contains():
    state = get_state_machine()

    assert "idle" in state
    assert "not in state" not in state


def test_state_machine_getitem():
    state = get_state_machine()

    assert state[0] == "idle"
    assert state[2] == "selected"


def test_state_machine_copy():
    og = get_state_machine()

    # Add transitions, keep states
    state = og.copy(
        add_transitions={
            "/": {
                "CLICKED": "/checked",
            },
            "/checked": {
                "CLICKED": "/",
            },
        },
    )
    assert state.states == state.states
    assert state._transitions == {
        **og._transitions,
        **{
            "/": {
                "CLICKED": "/checked",
                "SUBSTATE_ENTER_BLUR": "/blur",
            },
            "/checked": {
                "CLICKED": "/",
            },
            "/blur": {
                "SUBSTATE_EXIT_BLUR": "/",
            },
        },
    }

    # Overwrite transitions, keep states
    state = og.copy(
        transitions={
            "/": {
                "CLICKED": "/checked",
            },
            "/checked": {
                "CLICKED": "/",
            },
        },
    )
    assert state.states == state.states
    assert state._transitions == {
        "/": {
            "CLICKED": "/checked",
        },
        "/checked": {
            "CLICKED": "/",
        },
    }

    # Keep transitions, add states
    state = og.copy(
        add_states=("test",),
    )
    assert state.states == tuple((*og.states, "test"))
    assert state._transitions == og._transitions

    # Keep transitions, overwrite states
    state = og.copy(
        states=("test",),
    )
    assert state.states == ("test",)
    assert state._transitions == og._transitions


def test_state_machine_substates():
    state = get_state_machine()

    state.apply_action("SUBSTATE_ENTER_BLUR")
    assert state() == "idle/blur"

    state.apply_action("HOVERED")
    assert state() == "hover/blur"

    state.apply_action("CLICKED")
    assert state() == "active/blur"

    state.apply_action("SUBSTATE_EXIT_BLUR")
    assert state() == "active"

    assert not state.apply_action("SUBSTATE_THIS_WONT_WORK")


def test_state_machine_eq():
    state = get_state_machine()

    assert state().startswith("idle")

    state.apply_action("SUBSTATE_ENTER_BLUR")
    assert state().endswith("/blur")



from typing import Any

from .text import Text


class Link(Text):
    """A purpose-made Text object for hyperlinks."""

    rules = """
    Link:
        height: 1
    """

    def __init__(self, content: str, to: str | None = None, **widget_args: Any) -> None:
        if to is None:
            to = content.lower()

        super().__init__(f"[~/{to}]{content}[/~]", **widget_args)
